import os
import sys
from typing import Optional, Dict, List, Tuple
from collections import defaultdict

import numpy as np
from Orange.data import Table, Domain, Variable, DiscreteVariable, StringVariable
from Orange.widgets import widget
from Orange.widgets.settings import Setting
from Orange.widgets.utils.signals import Input, Output

from AnyQt.QtWidgets import (
    QApplication,
    QWidget,
    QComboBox,
    QSizePolicy,
)

# Importations conditionnelles pour refléter l'environnement Orange add-on
if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic  # noqa: F401
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import (
        apply_modification_from_python_file,
    )
else:
    from orangecontrib.AAIT.utils.import_uic import uic  # type: ignore  # noqa: F401
    from orangecontrib.AAIT.utils.initialize_from_ini import (  # type: ignore
        apply_modification_from_python_file,
    )


@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWFusionNN(widget.OWWidget):
    """Fusion NxM simplifiée avec gestion correcte des types Orange."""

    name = "Fusion NxM"
    description = "Merge two Tables on a key; add the non-matched rows."
    category = "AAIT - TOOLBOX"
    icon = "icons/owfusion_nm.png"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owfusion_nm.png"
    priority = 1095

    want_main_area = True
    want_control_area = False

    # Paramètres persistés
    key1_a: Optional[str] = Setting(None)
    key1_b: Optional[str] = Setting(None)

    class Inputs:
        data_a = Input("Table 1", Table)
        data_b = Input("Table 2", Table)

    class Outputs:
        data = Output("Data", Table)

    class Error(widget.OWWidget.Error):
        no_keys = widget.Msg("Sélectionnez une clé pour A et B.")
        fusion_error = widget.Msg("Erreur de fusion: {}")

    def __init__(self) -> None:
        super().__init__()

        self._data_a: Optional[Table] = None
        self._data_b: Optional[Table] = None

        # Charger UI
        self.gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owfusion_nm.ui")
        self.form: QWidget = uic.loadUi(self.gui)
        self.mainArea.layout().addWidget(self.form)
        self.form.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        self.cb_key1_a: QComboBox = getattr(self.form, "cb_key1_a")
        self.cb_key1_b: QComboBox = getattr(self.form, "cb_key1_b")
        self.cb_key1_a.currentIndexChanged.connect(self._on_key_changed)
        self.cb_key1_b.currentIndexChanged.connect(self._on_key_changed)

    @Inputs.data_a
    def set_data_a(self, data: Optional[Table]):
        self._data_a = data
        self._update_combos()
        self._apply_fusion()

    @Inputs.data_b
    def set_data_b(self, data: Optional[Table]):
        self._data_b = data
        self._update_combos()
        self._apply_fusion()

    @staticmethod
    def _get_var_names(table: Optional[Table]) -> List[str]:
        """Retourne les noms de toutes les variables d'une table."""
        if table is None:
            return []
        domain = table.domain
        names = [v.name for v in domain.attributes]
        names.extend(v.name for v in domain.class_vars)
        names.extend(v.name for v in domain.metas)
        seen = set()
        return [n for n in names if not (n in seen or seen.add(n))]

    @staticmethod
    def _get_column(table: Table, var_name: str) -> np.ndarray:
        """Extrait une colonne d'une table par nom de variable."""
        domain = table.domain
        for i, var in enumerate(domain.attributes):
            if var.name == var_name:
                return table.X[:, i]
        for i, var in enumerate(domain.class_vars):
            if var.name == var_name:
                if table.Y.ndim == 1:
                    return table.Y if i == 0 else np.full(len(table), np.nan)
                return table.Y[:, i]
        for i, var in enumerate(domain.metas):
            if var.name == var_name:
                return table.metas[:, i]
        return np.array([])

    @staticmethod
    def _find_var(table: Table, var_name: str) -> Tuple[Optional[Variable], str, int]:
        """Trouve une variable et retourne (variable, type, index)."""
        domain = table.domain
        for i, var in enumerate(domain.attributes):
            if var.name == var_name:
                return var, "attr", i
        for i, var in enumerate(domain.class_vars):
            if var.name == var_name:
                return var, "class", i
        for i, var in enumerate(domain.metas):
            if var.name == var_name:
                return var, "meta", i
        return None, "", -1

    @staticmethod
    def _is_valid_key(value) -> bool:
        """Vérifie si une valeur peut être utilisée comme clé de fusion."""
        if value is None:
            return False
        try:
            return not np.isnan(value)
        except (TypeError, ValueError):
            return True

    @staticmethod
    def _keys_equal(v1, v2) -> bool:
        """Compare deux valeurs de clés."""
        if not OWFusionNN._is_valid_key(v1) or not OWFusionNN._is_valid_key(v2):
            return False
        return v1 == v2

    def _update_combos(self) -> None:
        """Met à jour les comboboxes avec les noms de variables."""
        names_a = self._get_var_names(self._data_a)
        names_b = self._get_var_names(self._data_b)

        # Mise à jour combo A
        self.cb_key1_a.blockSignals(True)
        current_a = self.cb_key1_a.currentText()
        self.cb_key1_a.clear()
        self.cb_key1_a.addItems(names_a)
        if self.key1_a and self.key1_a in names_a:
            self.cb_key1_a.setCurrentText(self.key1_a)
        elif current_a in names_a:
            self.cb_key1_a.setCurrentText(current_a)
        elif names_a:
            self.cb_key1_a.setCurrentIndex(0)
        self.cb_key1_a.blockSignals(False)

        # Mise à jour combo B
        self.cb_key1_b.blockSignals(True)
        current_b = self.cb_key1_b.currentText()
        self.cb_key1_b.clear()
        self.cb_key1_b.addItems(names_b)
        if self.key1_b and self.key1_b in names_b:
            self.cb_key1_b.setCurrentText(self.key1_b)
        elif current_b in names_b:
            self.cb_key1_b.setCurrentText(current_b)
        elif names_b:
            self.cb_key1_b.setCurrentIndex(0)
        self.cb_key1_b.blockSignals(False)

    def _on_key_changed(self) -> None:
        """Appelé quand l'utilisateur change une clé."""
        self.key1_a = self.cb_key1_a.currentText() or None
        self.key1_b = self.cb_key1_b.currentText() or None
        self._apply_fusion()

    def _apply_fusion(self) -> None:
        """Effectue la fusion des tables."""
        # Vérifications préliminaires
        if self._data_a is None or self._data_b is None:
            self.Outputs.data.send(None)
            return

        self.Error.clear()

        if not self.key1_a or not self.key1_b:
            self.Error.no_keys()
            self.Outputs.data.send(None)
            return

        try:
            result = self._merge_tables(self._data_a, self._data_b, self.key1_a, self.key1_b)
            self.Outputs.data.send(result)
        except Exception as e:
            self.Error.fusion_error(str(e))
            self.Outputs.data.send(None)

    def _merge_tables(self, table_a: Table, table_b: Table, key_a: str, key_b: str) -> Table:
        """Fusionne deux tables sur leurs clés."""
        # Extraire les colonnes de clés
        col_a = self._get_column(table_a, key_a)
        col_b = self._get_column(table_b, key_b)

        # Construire l'index B: valeur -> liste d'indices
        index_b: Dict[any, List[int]] = defaultdict(list)
        for j, val in enumerate(col_b):
            if self._is_valid_key(val):
                try:
                    index_b[val].append(j)
                except TypeError:
                    # Non-hashable, on gérera avec fallback
                    pass

        # Trouver les correspondances
        matches: List[Tuple[int, int]] = []
        matched_a = np.zeros(len(col_a), dtype=bool)
        matched_b = np.zeros(len(col_b), dtype=bool)

        for i, val_a in enumerate(col_a):
            if not self._is_valid_key(val_a):
                continue

            # Chercher dans l'index
            indices_b = index_b.get(val_a)
            if indices_b is None:
                # Fallback: comparaison directe
                indices_b = [j for j, val_b in enumerate(col_b) if self._keys_equal(val_a, val_b)]

            for j in indices_b:
                matches.append((i, j))
                matched_a[i] = True
                matched_b[j] = True

        # Lignes non appariées
        only_a = [i for i, m in enumerate(matched_a) if not m]
        only_b = [j for j, m in enumerate(matched_b) if not m]

        # Construire le domaine de sortie
        out_domain = self._build_output_domain(table_a, table_b, key_a, key_b)

        # Construire les données de sortie
        n_rows = len(matches) + len(only_a) + len(only_b)
        X_out, Y_out, M_out = self._allocate_output_arrays(out_domain, n_rows)

        # Remplir les données
        row = 0

        # Lignes appariées (A + B)
        for i, j in matches:
            self._fill_row(X_out, Y_out, M_out, row, out_domain, table_a, table_b, i, j, key_a, key_b)
            row += 1

        # Lignes A uniquement
        for i in only_a:
            self._fill_row(X_out, Y_out, M_out, row, out_domain, table_a, table_b, i, None, key_a, key_b)
            row += 1

        # Lignes B uniquement
        for j in only_b:
            self._fill_row(X_out, Y_out, M_out, row, out_domain, table_a, table_b, None, j, key_a, key_b)
            row += 1

        # Créer la table de sortie
        result = Table.from_numpy(out_domain, X_out, Y_out, M_out)
        result.name = f"{table_a.name or 'A'} ⋈ {table_b.name or 'B'}"

        return result

    def _build_output_domain(self, table_a: Table, table_b: Table,
                            key_a: str, key_b: str) -> Domain:
        """Construit le domaine de sortie avec toutes les variables en StringVariable."""
        same_key = (key_a == key_b)

        # Collecter tous les noms pour détecter les overlaps
        names_a = set(self._get_var_names(table_a))
        names_b = set(self._get_var_names(table_b))
        overlap = names_a & names_b

        out_metas: List[Variable] = []

        # Ajouter variables de A (toutes en metas/string)
        for var in list(table_a.domain.attributes) + list(table_a.domain.class_vars) + list(table_a.domain.metas):
            if var.name in overlap and not (same_key and var.name == key_a):
                out_metas.append(StringVariable(var.name + "_A"))
            else:
                out_metas.append(StringVariable(var.name))

        # Ajouter variables de B (toutes en metas/string)
        for var in list(table_b.domain.attributes) + list(table_b.domain.class_vars) + list(table_b.domain.metas):
            if same_key and var.name == key_b:
                continue  # Éviter duplication de la clé
            if var.name in overlap:
                out_metas.append(StringVariable(var.name + "_B"))
            else:
                out_metas.append(StringVariable(var.name))

        # Tout en metas (pas d'attributes ni de class_vars)
        return Domain([], [], out_metas)

    def _rename_var(self, var: Variable, suffix: str) -> Variable:
        """Crée une nouvelle variable avec un suffixe (toujours en StringVariable)."""
        new_name = var.name + suffix
        # Toujours retourner une StringVariable pour éviter les problèmes de types
        return StringVariable(new_name)

    def _allocate_output_arrays(self, domain: Domain, n_rows: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Alloue les arrays de sortie (tout en metas/string)."""
        # Pas d'attributes ni de class_vars, tout est en metas
        X_out = np.empty((n_rows, 0), dtype=float)
        Y_out = np.empty((n_rows, 0), dtype=float)
        M_out = np.full((n_rows, len(domain.metas)), "", dtype=object)

        return X_out, Y_out, M_out

    def _fill_row(self, X_out: np.ndarray, Y_out: np.ndarray, M_out: np.ndarray,
                  row_idx: int, out_domain: Domain,
                  table_a: Table, table_b: Table,
                  idx_a: Optional[int], idx_b: Optional[int],
                  key_a: str, key_b: str) -> None:
        """Remplit une ligne de sortie avec les données de A et/ou B (tout en metas/string)."""
        same_key = (key_a == key_b)

        # Tout est en metas maintenant
        for out_idx, out_var in enumerate(out_domain.metas):
            value = self._get_var_value(out_var, table_a, table_b, idx_a, idx_b, key_a, key_b, same_key, "meta")
            # Convertir en string
            if value is None or (isinstance(value, float) and np.isnan(value)):
                M_out[row_idx, out_idx] = ""
            elif isinstance(value, (list, tuple, np.ndarray)):
                M_out[row_idx, out_idx] = str(value)
            else:
                M_out[row_idx, out_idx] = str(value)

    def _get_var_value(self, out_var: Variable, table_a: Table, table_b: Table,
                      idx_a: Optional[int], idx_b: Optional[int],
                      key_a: str, key_b: str, same_key: bool, var_type: str):
        """Récupère la valeur d'une variable depuis la table source appropriée."""
        var_name = out_var.name

        # Déterminer si c'est une variable de A ou B
        if var_name.endswith("_A"):
            source_name = var_name[:-2]
            table, idx = table_a, idx_a
        elif var_name.endswith("_B"):
            source_name = var_name[:-2]
            table, idx = table_b, idx_b
        else:
            # Pas de suffixe: chercher d'abord dans A, puis dans B
            source_name = var_name
            var_a, _, _ = self._find_var(table_a, source_name)
            if var_a is not None:
                table, idx = table_a, idx_a
            else:
                table, idx = table_b, idx_b

        # Si pas d'index, retourner None
        if idx is None:
            return None

        # Extraire la valeur
        var, vtype, vidx = self._find_var(table, source_name)
        if var is None:
            return None

        # Récupérer la valeur brute
        raw_value = None
        if vtype == "attr":
            raw_value = table.X[idx, vidx]
        elif vtype == "class":
            if table.Y.ndim == 1:
                raw_value = table.Y[idx] if vidx == 0 else np.nan
            else:
                raw_value = table.Y[idx, vidx]
        elif vtype == "meta":
            raw_value = table.metas[idx, vidx]

        # Convertir les variables discrètes en leurs noms de valeurs
        if isinstance(var, DiscreteVariable) and raw_value is not None:
            try:
                if not np.isnan(raw_value):
                    value_idx = int(raw_value)
                    if 0 <= value_idx < len(var.values):
                        return var.values[value_idx]
            except (ValueError, TypeError):
                pass

        return raw_value


if __name__ == "__main__":
    app = QApplication(sys.argv)
    w = OWFusionNN()
    w.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()



import os
import re
from typing import Any, Iterable, List, Set
from pygments.formatters import HtmlFormatter
import markdown
import sys
# Dans la classe OWCorpusViewer, ajoutez les méthodes suivantes :
import numpy as np
from AnyQt.QtCore import (
    QAbstractListModel,
    QEvent,
    QItemSelection,
    QItemSelectionModel,
    QItemSelectionRange,
    QModelIndex,
    QSortFilterProxyModel,
    Qt,
    QUrl,
)
from AnyQt.QtWidgets import (
    QAbstractItemView,
    QApplication,
    QHeaderView,
    QListView,
    QSizePolicy,
    QSplitter,
    QTableView,
)
from Orange.data import Variable
from Orange.data.domain import Domain, filter_visible
from Orange.widgets import gui
from Orange.widgets.settings import ContextSetting, Setting, DomainContextHandler
from Orange.widgets.utils.annotated_data import create_annotated_table
from Orange.widgets.utils.concurrent import ConcurrentWidgetMixin, TaskState
from Orange.widgets.utils.itemmodels import DomainModel
from Orange.widgets.widget import Input, Msg, Output, OWWidget
from orangecanvas.gui.utils import disconnected
from orangewidget.utils.listview import ListViewSearch

from orangecontrib.text.corpus import Corpus
if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils import widget_positioning
else:
    from orangecontrib.AAIT.utils import widget_positioning


HTML = """
<!doctype html>
    <html>
    <head>
    <meta charset='utf-8'>
    <style>
    /* Votre CSS existant */
    table {{ border-collapse: collapse; }}
    mark {{ background: #FFCD28; }}

    tr > td {{
        padding-bottom: 3px;
        padding-top: 3px;
    }}

    body {{
        font-family: Helvetica;
        font-size: 10pt;
    }}

    .line {{ border-bottom: 1px solid #000; }}
    .separator {{ height: 5px; }}

    .variables {{
        vertical-align: top;
        padding-right: 10px;
    }}

    .content {{
        overflow-wrap: break-word;
        word-wrap: break-word;
        -ms-word-break: break-all;
        word-break: break-all;
        word-break: break-word;
        -ms-hyphens: auto;
        -moz-hyphens: auto;
        -webkit-hyphens: auto;
        hyphens: auto;
    }}

    .token {{
        padding: 3px;
        border: 1px #B0B0B0 solid;
        margin-right: 5px;
        margin-bottom: 5px;
        display: inline-block;
    }}

    img {{
        max-width: 100%;
    }}
    </style>
    {additional_css}
    </head>
    <body>
    {html_content}
    </body>
    </html>
    """

SEPARATOR = (
    '<tr class="line separator"><td/><td/></tr>'
    '<tr class="separator"><td/><td/></tr>'
)


def _count_matches(content: List[str], search_string: str, state: TaskState) -> int:
    """
    Count number of appears of any terms in search_string in content texts.

    Parameters
    ----------
    content
        List of texts where we count appearances
    search_string
        Strings that are searched in texts. This parameter has a format
        term1|term2|term3|...

    Returns
    -------
    Number of all matches of search_string in all texts in content list
    """
    matches = 0
    if search_string:
        regex = re.compile(search_string.strip("|"), re.IGNORECASE)
        for i, text in enumerate(content):
            matches += len(regex.findall(text))
            state.set_progress_value((i + 1) / len(content) * 100)
    return matches

def is_qt_offscreen():
    return os.environ.get('QT_QPA_PLATFORM') == 'offscreen'

class DocumentListModel(QAbstractListModel):
    """
    Custom model for listing documents. Using custom model since Onrage's
    pylistmodel is too slow for large number of documents
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__visible_data = []
        self.__filter_content = []

    def data(self, index: QModelIndex, role: int = Qt.DisplayRole) -> Any:
        if role == Qt.DisplayRole:
            return self.__visible_data[index.row()]
        elif role == Qt.UserRole:
            return self.__filter_content[index.row()]

    def rowCount(self, parent: QModelIndex = None, *args, **kwargs) -> int:
        return len(self.__visible_data)

    def setup_data(self, data: List[str], content: List[str]):
        self.beginResetModel()
        self.__visible_data = data
        self.__filter_content = content
        self.endResetModel()

    def update_filter_content(self, content: List[str]):
        assert len(content) == len(self.__visible_data)
        self.__filter_content = content

    def get_filter_content(self) -> List[str]:
        return self.__filter_content

    def clear(self):
        self.beginResetModel()
        self.__visible_data = []
        self.__filter_content = []
        self.endResetModel()


class DocumentsFilterProxyModel(QSortFilterProxyModel):
    """Filter model for documents list"""

    __regex = None

    def set_filter_string(self, filter_string: str):
        self.__regex = re.compile(filter_string.strip("|"), re.IGNORECASE)
        self.invalidateFilter()

    def filterAcceptsRow(self, source_row: int, source_parent: QModelIndex) -> bool:
        """Filter document that mathc the filter string"""
        if self.__regex is None:
            # filter is not defined yet - show all
            return True
        else:
            index = self.sourceModel().index(source_row, 0, source_parent)
            content = self.sourceModel().data(index, Qt.UserRole)
            res = self.__regex.search(content)
            return bool(res)


class DocumentTableView(QTableView):
    """TableView that disables unselecting all items"""

    def selectionCommand(
        self, index: QModelIndex, event: QEvent = None
    ) -> QItemSelectionModel.SelectionFlags:
        flags = super().selectionCommand(index, event)
        selmodel = self.selectionModel()
        if not index.isValid():  # Click on empty viewport; don't clear
            return QItemSelectionModel.NoUpdate
        if selmodel.isSelected(index):
            currsel = selmodel.selectedIndexes()
            if len(currsel) == 1 and index == currsel[0]:
                # Is the last selected index; do not deselect it
                return QItemSelectionModel.NoUpdate
        if (
            event is not None
            and event.type() == QEvent.MouseMove
            and flags & QItemSelectionModel.ToggleCurrent
        ):
            # Disable ctrl drag 'toggle'; can be made to deselect the last
            # index, would need to keep track of the current selection
            # (selectionModel does this but does not expose it)
            flags &= ~QItemSelectionModel.Toggle
            flags |= QItemSelectionModel.Select
        return flags


class VariableListViewSearch(ListViewSearch):
    """ListViewSearch that disables unselecting all items in the list"""

    def selectionCommand(
        self, index: QModelIndex, event: QEvent = None
    ) -> QItemSelectionModel.SelectionFlags:
        flags = super().selectionCommand(index, event)
        selmodel = self.selectionModel()
        if not index.isValid():  # Click on empty viewport; don't clear
            return QItemSelectionModel.NoUpdate
        if selmodel.isSelected(index):
            currsel = selmodel.selectedIndexes()
            if len(currsel) == 1 and index == currsel[0]:
                # Is the last selected index; do not deselect it
                return QItemSelectionModel.NoUpdate
        if (
            event is not None
            and event.type() == QEvent.MouseMove
            and flags & QItemSelectionModel.ToggleCurrent
        ):
            # Disable ctrl drag 'toggle'; can be made to deselect the last
            # index, would need to keep track of the current selection
            # (selectionModel does this but does not expose it)
            flags &= ~QItemSelectionModel.Toggle
            flags |= QItemSelectionModel.Select
        return flags

    def set_selection(self, items: Iterable[Variable]):
        """Set selected items in the list view"""
        model = self.model()
        values = self.model()[:]
        items = [it for it in items if it in values]
        selection = QItemSelection()
        if items:
            for val in items:
                index = values.index(val)
                selection.merge(
                    QItemSelection(model.index(index, 0), model.index(index, 0)),
                    QItemSelectionModel.Select,
                )
        self.selectionModel().select(selection, QItemSelectionModel.ClearAndSelect)


class VisibleDomainModel(DomainModel):
    """Domain model that filter only visible features"""

    def set_domain(self, domain):
        if domain is not None:
            domain = Domain(
                filter_visible(domain.attributes),
                class_vars=filter_visible(domain.class_vars),
                metas=filter_visible(domain.metas),
            )
        super().set_domain(domain)


class OWCorpusViewer(OWWidget, ConcurrentWidgetMixin):
    name = "MarkDown Viewer"
    description = "Display corpus contents with Markdown."
    icon = "icons/MDViewer.png"
    category = "AAIT - TOOLBOX"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/MDViewer.png"
    priority = 1002
    keywords = "Markdown viewer, display"
    #save_position = False



    class Inputs:
        corpus = Input("Corpus", Corpus, replaces=["Data"])
        input_autoshow = Input("AutoShowConfiguration", str, auto_summary=False)
    class Outputs:
        matching_docs = Output("Matching Docs", Corpus, default=True)
        other_docs = Output("Other Docs", Corpus)
        corpus = Output("Corpus", Corpus)


    str_WidgetPositionning: str=Setting("None")
    settingsHandler = DomainContextHandler()

    settings_version = 2
    search_features: List[Variable] = ContextSetting([])
    display_features: List[Variable] = ContextSetting([])
    selected_documents: Set[int] = Setting({0}, schema_only=True)
    regexp_filter = ContextSetting("")
    show_tokens = Setting(False)
    autocommit = Setting(True)

    class Warning(OWWidget.Warning):
        no_feats_search = Msg("No features included in search.")
        no_feats_display = Msg("No features selected for display.")

    def __init__(self):
        super().__init__()
        ConcurrentWidgetMixin.__init__(self)

        self.corpus = None  # Corpus
        self.__pending_selected_documents = self.selected_documents

        # Info attributes
        self.update_info()
        info_box = gui.widgetBox(self.controlArea, "Info")
        gui.label(info_box, self, "Tokens: %(n_tokens)s")
        gui.label(info_box, self, "Types: %(n_types)s")
        gui.label(info_box, self, "Matching documents: %(n_matching)s")
        gui.label(info_box, self, "Matches: %(n_matches)s")

        # Search features
        ex_sel = QListView.ExtendedSelection
        search_box = gui.widgetBox(self.controlArea, "Search features")
        self.search_listbox = sl = VariableListViewSearch(selectionMode=ex_sel)
        search_box.layout().addWidget(sl)
        sl.setModel(VisibleDomainModel(separators=False))
        sl.selectionModel().selectionChanged.connect(self.search_features_changed)

        # Display features
        display_box = gui.widgetBox(self.controlArea, "Display features")
        self.display_listbox = dl = VariableListViewSearch(selectionMode=ex_sel)
        display_box.layout().addWidget(dl)
        dl.setModel(VisibleDomainModel(separators=False))
        dl.selectionModel().selectionChanged.connect(self.display_features_changed)

        self.show_tokens_checkbox = gui.checkBox(
            display_box,
            self,
            "show_tokens",
            "Show Tokens && Tags",
            callback=self.show_docs,
        )

        # Auto-commit box
        gui.auto_commit(
            self.controlArea, self, "autocommit", "Send data", "Auto send is on"
        )

        # Search
        self.filter_input = gui.lineEdit(
            self.mainArea,
            self,
            "regexp_filter",
            orientation=Qt.Horizontal,
            sizePolicy=QSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Fixed),
            label="RegExp Filter:",
            callback=self.refresh_search,
        )

        # Main area
        self.splitter = QSplitter(orientation=Qt.Horizontal, childrenCollapsible=False)
        # Document list
        self.doc_list = DocumentTableView()
        self.doc_list.setSelectionBehavior(QTableView.SelectRows)
        self.doc_list.setSelectionMode(QTableView.ExtendedSelection)
        self.doc_list.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.doc_list.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)
        self.doc_list.horizontalHeader().setVisible(False)
        self.splitter.addWidget(self.doc_list)

        self.doc_list_model = DocumentListModel()
        proxy_model = DocumentsFilterProxyModel()
        proxy_model.setSourceModel(self.doc_list_model)
        self.doc_list.setModel(proxy_model)
        self.doc_list.selectionModel().selectionChanged.connect(self.selection_changed)
        # Document contents
        self.doc_webview = None
        if not is_qt_offscreen():
            self.doc_webview = gui.WebviewWidget(self.splitter, debug=False)
            self.mainArea.layout().addWidget(self.splitter)
        widget_positioning.show_and_adjust_at_opening(self,str(self.str_WidgetPositionning))


    def copy_to_clipboard(self):
        if self.doc_webview is not None:
            text = self.doc_webview.selectedText()
            QApplication.clipboard().setText(text)
        else:
            QApplication.clipboard().setText("")

    @Inputs.corpus
    def set_data(self, corpus=None):
        self.closeContext()
        self.reset_widget()
        self.corpus = corpus
        if corpus is not None:
            self.setup_controls()
            self.openContext(self.corpus)
            self.doc_list.model().set_filter_string(self.regexp_filter)
            self.select_variables()
            self.list_docs()
            self.update_info()
            self.set_selection()
            self.show_docs()
        self.commit.now()

    @Inputs.input_autoshow
    def set_input_autoshow(self, le_str):
        if le_str is not None:
            self.str_WidgetPositionning=str(le_str)




    def reset_widget(self):
        # Corpus
        self.corpus = None
        # Widgets
        self.search_listbox.model().set_domain(None)
        self.display_listbox.model().set_domain(None)
        self.filter_input.clear()
        self.update_info()
        # Models/vars
        self.doc_list_model.clear()
        # Warnings
        self.Warning.clear()
        # WebView
        if self.doc_webview is not None:
            self.doc_webview.setHtml("")

    def setup_controls(self):
        """Setup controls in control area"""
        domain = self.corpus.domain
        if not self.corpus.has_tokens():
            self.show_tokens_checkbox.setCheckState(Qt.Unchecked)
        self.show_tokens_checkbox.setEnabled(self.corpus.has_tokens())
        self.search_listbox.model().set_domain(domain)
        self.display_listbox.model().set_domain(domain)
        self.search_features = self.search_listbox.model()[:]
        self.display_features = self.display_listbox.model()[:]

    def select_variables(self):
        """Set selection to display and search features view boxes"""
        smodel = self.search_listbox.model()
        dmodel = self.display_listbox.model()
        # it can happen that domain handler will set some features that are
        # not part of domain - remove them
        self.search_features = [f for f in self.search_features if f in smodel]
        self.display_features = [f for f in self.display_features if f in dmodel]
        # if no features after removing non-existent, select all - default
        if not self.search_features:
            self.search_features = smodel[:]
        if not self.display_features:
            self.display_features = dmodel[:]
        with disconnected(
            self.search_listbox.selectionModel().selectionChanged,
            self.search_features_changed,
        ):
            self.search_listbox.set_selection(self.search_features)
        with disconnected(
            self.display_listbox.selectionModel().selectionChanged,
            self.display_features_changed,
        ):
            self.display_listbox.set_selection(self.display_features)

    def list_docs(self):
        """List documents into the left scrolling area"""
        docs = self.regenerate_docs()
        self.doc_list_model.setup_data(self.corpus.titles.tolist(), docs)

    def get_selected_indexes(self) -> Set[int]:
        m = self.doc_list.model().mapToSource
        return {m(i).row() for i in self.doc_list.selectionModel().selectedRows()}

    def set_selection(self) -> None:
        """
        Select documents in selected_documents attribute in the view
        """
        self.selected_documents = self.__pending_selected_documents
        self.__pending_selected_documents = {0}
        view = self.doc_list
        model = view.model()
        source_model = model.sourceModel()

        selection = QItemSelection()
        self.selected_documents = {
            r for r in self.selected_documents if r < len(self.corpus)
        }
        for row in self.selected_documents:
            index = model.mapFromSource(source_model.index(row, 0))
            selection.append(QItemSelectionRange(index, index))
        # don't emit selection change to avoid double call of commit function
        # it is already called from set_data
        with disconnected(
            self.doc_list.selectionModel().selectionChanged, self.selection_changed
        ):
            view.selectionModel().select(selection, QItemSelectionModel.ClearAndSelect)

    def selection_changed(self) -> None:
        """Function is called every time the selection changes"""
        self.selected_documents = self.get_selected_indexes()
        self.show_docs()
        self.commit.deferred()


    def __mark_text(self, text):
        search_keyword = self.regexp_filter.strip("|")
        if not search_keyword:
            return text

        try:
            reg = re.compile(search_keyword, re.IGNORECASE | re.MULTILINE)
        except re.error:
            return text

        matches = list(reg.finditer(text))
        if not matches:
            return text

        text = list(text)
        for m in matches[::-1]:
            text[m.start() : m.end()] = list(
                f'<mark data-markjs="true">{"".join(text[m.start(): m.end()])}</mark>'
            )
        return "".join(text)

    @staticmethod
    def __get_selected_rows(view: QListView) -> List[Variable]:
        rows = view.selectionModel().selectedRows()
        values = view.model()[:]
        return [values[row.row()] for row in sorted(rows, key=lambda r: r.row())]

    def search_features_changed(self):
        self.search_features = self.__get_selected_rows(self.search_listbox)
        if self.corpus:
            self.doc_list_model.update_filter_content(self.regenerate_docs())
        self.doc_list.model().invalidateFilter()
        self.refresh_search()

    def display_features_changed(self):
        self.display_features = self.__get_selected_rows(self.display_listbox)
        self.show_docs()

    def regenerate_docs(self) -> List[str]:
        self.Warning.no_feats_search.clear()
        if len(self.search_features) == 0:
            self.Warning.no_feats_search()
        return self.corpus.documents_from_features(self.search_features)

    def refresh_search(self):
        if self.corpus is not None:
            self.doc_list.model().set_filter_string(self.regexp_filter)
            if not self.selected_documents:
                # when currently selected items are filtered selection is empty
                # select first element in the view in that case
                self.doc_list.setCurrentIndex(self.doc_list.model().index(0, 0))
            self.update_info()
            self.start(
                _count_matches,
                self.doc_list_model.get_filter_content(),
                self.regexp_filter,
            )
            self.show_docs()
            self.commit.deferred()

    def on_done(self, res: int):
        """When matches count is done show the result in the label"""
        self.n_matches = res if res is not None else "n/a"

    def on_exception(self, ex):
        raise ex

    def update_info(self):
        if self.corpus is not None:
            has_tokens = self.corpus.has_tokens()
            self.n_matching = f"{self.doc_list.model().rowCount()}/{len(self.corpus)}"
            self.n_tokens = self.corpus.count_tokens() if has_tokens else "n/a"
            self.n_types = self.corpus.count_unique_tokens() if has_tokens else "n/a"
        else:
            self.n_matching = "n/a"
            self.n_matches = "n/a"
            self.n_tokens = "n/a"
            self.n_types = "n/a"

    @gui.deferred
    def commit(self):
        matched = unmatched = annotated_corpus = None
        if self.corpus is not None:
            selected_docs = sorted(self.get_selected_indexes())
            matched = self.corpus[selected_docs] if selected_docs else None
            mask = np.ones(len(self.corpus), bool)
            mask[selected_docs] = 0
            unmatched = self.corpus[mask] if mask.any() else None
            annotated_corpus = create_annotated_table(self.corpus, selected_docs)
        self.Outputs.matching_docs.send(matched)
        self.Outputs.other_docs.send(unmatched)
        self.Outputs.corpus.send(annotated_corpus)

    def send_report(self):
        self.report_items(
            (
                ("Query", self.regexp_filter),
                ("Matching documents", self.n_matching),
                ("Matches", self.n_matches),
            )
        )

    def showEvent(self, event):
        super().showEvent(event)
        self.update_splitter()

    def update_splitter(self):
        """
        Update splitter that document list on the left never take more
        than 1/3 of the space. It is only set on showEvent. If user
        later changes sizes it stays as it is.
        """
        w1, w2 = self.splitter.sizes()
        ws = w1 + w2
        if w2 < 2 / 3 * ws:
            self.splitter.setSizes([int(ws * 1 / 3), int(ws * 2 / 3)])

    @classmethod
    def migrate_context(cls, context, version):
        if version < 2:
            f_order = context.values.pop("display_features", None)
            display_idx = context.values.pop("display_indices", [])
            search_ids = context.values.pop("search_indices", [])
            if f_order is not None:
                f_order = f_order[0]
                display_features = [f_order[i] for i in display_idx if i < len(f_order)]
                search_features = [f_order[i] for i in search_ids if i < len(f_order)]
                context.values["display_features"] = (display_features, -3)
                context.values["search_features"] = (search_features, -3)

            # old widget used PerfectDomainContextHandler with MATCH_VALUES_ALL
            # now it uses DomainContextHandler. The difference are:
            # - perfect handler stores values in tuple while domain in dicts
            # - domain context handler store class_vars together with attributes
            #   while perfect handler store them separately
            # - since MATCH_VALUES_ALL was used discrete var values were stored
            #   with var name (replacing them with id for discrete var - 1)
            if hasattr(context, "class_vars"):
                context.attributes = {
                    attr: 1 if isinstance(v, list) else v
                    for attr, v in context.attributes + context.class_vars
                }
                context.metas = dict(context.metas)
                delattr(context, "class_vars")

    def convert_markdown_with_syntax_highlighting(self, markdown_text, base_path=""):
        """
        Convertit le texte Markdown avec syntaxe de code colorée en HTML.
        """

        # --- Patch : normaliser les listes pour que les puces après un "1." restent des sous-listes ---
        markdown_text = self._normalize_lists_for_markdown(markdown_text)
        # ---------------------------------------------------------------------

        # Convertir le Markdown en HTML avec extensions pour le code
        html_content = markdown.markdown(
            markdown_text,
            extensions=['fenced_code', 'codehilite', 'tables', 'toc', 'md_in_html'],
            extension_configs={
                'codehilite': {
                    'guess_lang': False,
                    'pygments_style': 'default',
                    'noclasses': False  # Utiliser les classes CSS
                }
            }
        )

        # Ajouter le CSS pour Pygments
        css = HtmlFormatter().get_style_defs('.codehilite')
        html_content = f"<style>{css}</style>{html_content}"

        # Corriger les chemins d'images pour les fichiers locaux
        html_content = self.adjust_image_paths(html_content, base_path)
        return html_content

    def _normalize_lists_for_markdown(self, text: str) -> str:
        """
        Transforme les puces qui suivent immédiatement un item numéroté
        en sous-liste (en les indentant), pour éviter une renumérotation 1,2,3,4,5.

        Exemple :
            1. Titre
            2. Autre titre
            - sous-étape
            - sous-étape

        devient :
            1. Titre
            2. Autre titre
                - sous-étape
                - sous-étape
        """
        lines = text.splitlines()
        out = []
        in_ordered_list = False

        for line in lines:
            stripped = line.lstrip()

            # Ligne d'une liste numérotée : "1. xxx"
            if re.match(r'^\d+\.\s+', stripped):
                in_ordered_list = True
                # On normalise en supprimant l'indentation superflue éventuelle
                out.append(stripped)
                continue

            # Ligne d'une puce qui suit une liste numérotée
            if in_ordered_list and re.match(r'^[-*+]\s+', stripped):
                # Si elle n'est pas déjà indentée, on l'indente comme sous-liste
                if not line.startswith('    '):
                    out.append('    ' + stripped)
                else:
                    out.append(line)
                continue

            # Autre contenu : on sort du contexte "liste numérotée"
            in_ordered_list = False
            out.append(line)

        return "\n".join(out)

    def adjust_image_paths(self, html_content, base_path):
        """
        Corrige les chemins d'images pour qu'ils soient valides même pour des chemins relatifs.
        """

        def replace_path(match):
            image_path = match.group(1)

            # Si l'image est locale, corriger le chemin
            if not image_path.startswith(('http://', 'https://')):
                full_path = os.path.abspath(os.path.join(base_path, image_path))
                if os.path.exists(full_path):
                    print(f"[INFO] Image trouvée : {full_path}")
                    return f'src="file:///{full_path}"'
                else:
                    print(f"[ERREUR] Image introuvable : {full_path}")
                    return f'src="{image_path}"'  # Affiche toujours le chemin, même s'il est incorrect
            else:
                print(f"[INFO] Image distante : {image_path}")
                return f'src="{image_path}"'

        # Remplacer les chemins d'image
        adjusted_html = re.sub(r'src="([^"]+)"', replace_path, html_content)
        return adjusted_html

    def is_markdown(self, text):

        """Détecte si le texte est probablement du Markdown.

        Returns :
            bool : le text balisé
            """

        return bool(re.search(r'(```.+?```)|(`.+?`)', text, re.DOTALL))

    def get_pygments_css(self):
        """Génère le CSS pour Pygments.
                Returns :
                    str : le CSS de Pygments.
                    """
        formatter = HtmlFormatter(style='default')
        return f"<style>{formatter.get_style_defs()}</style>"
    def show_docs(self):
        """Affiche les documents sélectionnés dans la zone de droite.

        Affiche les documents stockés dans `selected_documents` dans la zone
        de droite. Pour cela, elle construit un tableau HTML avec les valeurs
        des différentes colonnes de la base de données.
        """
        if self.corpus is None:
            return

        # Afficher un message d'avertissement si on n'a pas de colonnes
        # sélectionnées pour l'affichage
        self.Warning.no_feats_display.clear()
        if len(self.display_features) == 0:
            self.Warning.no_feats_display()

        parts = []
        for doc_count, c_index in enumerate(sorted(self.selected_documents)):
            text = ""
            for feature in self.display_features:
                # Récupérer la valeur de la colonne
                value = str(self.corpus[c_index, feature.name])


                if feature.attributes.get("type", "") == "image" and value != "?":
                    value = os.path.join(feature.attributes.get("origin", ""), value)
                    value = f'<img src="{value}"></img>'

                else:
                    value = self.convert_markdown_with_syntax_highlighting(value, base_path="")


                # Ajouter la valeur formattée à la colonne
                text += (
                    f'<tr><td class="variables"><strong>{feature.name}:</strong></td>'
                    f'<td class="content">{value}</td></tr>'
                )

            # Ajouter la colonne formattée à la liste des colonnes
            parts.append(text)

        # Concatener les colonnes pour obtenir un tableau HTML
        joined = SEPARATOR.join(parts)

        # Générer le code HTML complet avec le style CSS et le tableau
        html_content = f"<table>{joined}</table>"
        css_styles = self.get_pygments_css()
        base = QUrl.fromLocalFile(__file__)
        if self.doc_webview is not None:
            self.doc_webview.setHtml(
                HTML.format(
                    additional_css=css_styles,
                    html_content=html_content
                ),
                base
            )

if __name__ == "__main__":
    app = QApplication(sys.argv)
    obj = OWCorpusViewer()
    obj.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()


import os
from AnyQt.QtWidgets import QPushButton, QApplication, QRadioButton, QCheckBox,QSpinBox, QLineEdit

from Orange.widgets.settings import Setting
from Orange.widgets.widget import OWWidget, Output, Input
from Orange.data import Table
if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils import OperationSystem
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
    from Orange.widgets.orangecontrib.AAIT.utils import thread_management
    from Orange.widgets.orangecontrib.AAIT.utils.unlink_table_domain import unlink_domain
else:
    from orangecontrib.AAIT.utils import OperationSystem
    from orangecontrib.AAIT.utils.import_uic import uic
    from orangecontrib.AAIT.utils import thread_management
    from orangecontrib.AAIT.utils.unlink_table_domain import unlink_domain



class OwOperationSystem(OWWidget):
    name = "Operation System"
    description = "Basic operation on file and folder: delete copy/paste etc"
    icon = "icons/operationSystem.png"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/operationSystem.png"
    category = "AAIT - TOOLBOX"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/ow_OperationSystem.ui")
    priority = 10
    want_control_area = False

    class Inputs:
        input_table = Input("in data", Table)

    class Outputs:
        data = Output("out data", Table)

    str_spinBox: str = Setting("1")
    str_autosend: str = Setting("True")
    str_radio_box_checked=Setting("0")
    str_proxy: str = Setting("")

    @Inputs.input_table
    def input_table(self, data):
        self.error("")
        self.in_data = data
        if data is not None:
            if self.str_autosend=="True":
                self.run()


    def update_setting_from_qt_view(self):
        self.str_spinBox=str(self.spinBox.value())
        if self.checkBox_autosend.isChecked():
            self.str_autosend="True"
        else:
            self.str_autosend="False"
        if self.radioButton_deleteFoldersFiles.isChecked():
            self.str_radio_box_checked="0"
        if self.radioButton_createFolders.isChecked():
            self.str_radio_box_checked="1"
        if self.radioButton_moveRenamme.isChecked():
            self.str_radio_box_checked="2"
        if self.radioButton_copy.isChecked():
            self.str_radio_box_checked="3"
        if self.radioButton_sleep.isChecked():
            self.str_radio_box_checked="4"
        if self.radioButton_setProxy.isChecked():
            self.str_radio_box_checked="5"
        if self.radioButton_unlink_all_variable.isChecked():
            self.str_radio_box_checked="6"


    def __init__(self):
        super().__init__()
        self.selected_path = ""
        self.in_data = None
        self.thread = None

        self.setFixedWidth(700)
        self.setFixedHeight(400)
        uic.loadUi(self.gui, self)
        self.pushButton=self.findChild(QPushButton, 'pushButton')
        self.radioButton_deleteFoldersFiles=self.findChild(QRadioButton, 'radioButton_deleteFoldersFiles')
        self.radioButton_createFolders = self.findChild(QRadioButton, 'radioButton_createFolders')
        self.radioButton_moveRenamme = self.findChild(QRadioButton, 'radioButton_moveRenamme')
        self.radioButton_copy = self.findChild(QRadioButton, 'radioButton_copy')
        self.radioButton_sleep = self.findChild(QRadioButton, 'radioButton_sleep')
        self.radioButton_setProxy = self.findChild(QRadioButton, 'radioButton_setProxy')
        self.spinBox = self.findChild(QSpinBox, 'spinBox')
        self.checkBox_autosend= self.findChild(QCheckBox, 'checkBox_autosend')
        self.radioButton_unlink_all_variable=self.findChild(QRadioButton, 'radioButton_unlink_all_variable')
        self.spinBox.setValue(int(self.str_spinBox))

        self.edit_proxy = self.findChild(QLineEdit, 'lineEdit')
        self.edit_proxy.setPlaceholderText("")
        self.edit_proxy.setText(self.str_proxy)
        self.edit_proxy.editingFinished.connect(self.update_parameters)

        if self.str_autosend=="True":
            self.checkBox_autosend.setChecked(True)
        else:
            self.checkBox_autosend.setChecked(False)
        if self.str_radio_box_checked=="0":
            self.radioButton_deleteFoldersFiles.setChecked(True)
            self.radioButton_createFolders.setChecked(False)
            self.radioButton_moveRenamme.setChecked(False)
            self.radioButton_copy.setChecked(False)
            self.radioButton_sleep.setChecked(False)
            self.radioButton_setProxy.setChecked(False)
            self.radioButton_unlink_all_variable.setChecked(False)

        if self.str_radio_box_checked=="1":
            self.radioButton_deleteFoldersFiles.setChecked(False)
            self.radioButton_createFolders.setChecked(True)
            self.radioButton_moveRenamme.setChecked(False)
            self.radioButton_copy.setChecked(False)
            self.radioButton_sleep.setChecked(False)
            self.radioButton_setProxy.setChecked(False)
            self.radioButton_unlink_all_variable.setChecked(False)

        if self.str_radio_box_checked=="2":
            self.radioButton_deleteFoldersFiles.setChecked(False)
            self.radioButton_createFolders.setChecked(False)
            self.radioButton_moveRenamme.setChecked(True)
            self.radioButton_copy.setChecked(False)
            self.radioButton_sleep.setChecked(False)
            self.radioButton_setProxy.setChecked(False)
            self.radioButton_unlink_all_variable.setChecked(False)

        if self.str_radio_box_checked=="3":
            self.radioButton_deleteFoldersFiles.setChecked(False)
            self.radioButton_createFolders.setChecked(False)
            self.radioButton_moveRenamme.setChecked(False)
            self.radioButton_copy.setChecked(True)
            self.radioButton_sleep.setChecked(False)
            self.radioButton_setProxy.setChecked(False)
            self.radioButton_unlink_all_variable.setChecked(False)

        if self.str_radio_box_checked == "4":
            self.radioButton_deleteFoldersFiles.setChecked(False)
            self.radioButton_createFolders.setChecked(False)
            self.radioButton_moveRenamme.setChecked(False)
            self.radioButton_copy.setChecked(False)
            self.radioButton_sleep.setChecked(True)
            self.radioButton_setProxy.setChecked(False)
            self.radioButton_unlink_all_variable.setChecked(False)

        if self.str_radio_box_checked == "5":
            self.radioButton_deleteFoldersFiles.setChecked(False)
            self.radioButton_createFolders.setChecked(False)
            self.radioButton_moveRenamme.setChecked(False)
            self.radioButton_copy.setChecked(False)
            self.radioButton_sleep.setChecked(False)
            self.radioButton_setProxy.setChecked(True)
            self.radioButton_unlink_all_variable.setChecked(False)

        if self.str_radio_box_checked == "6":
            self.radioButton_deleteFoldersFiles.setChecked(False)
            self.radioButton_createFolders.setChecked(False)
            self.radioButton_moveRenamme.setChecked(False)
            self.radioButton_copy.setChecked(False)
            self.radioButton_sleep.setChecked(False)
            self.radioButton_setProxy.setChecked(False)
            self.radioButton_unlink_all_variable.setChecked(True)

        self.radioButton_deleteFoldersFiles.clicked.connect(self.update_setting_from_qt_view)
        self.radioButton_createFolders.clicked.connect(self.update_setting_from_qt_view)
        self.radioButton_moveRenamme.clicked.connect(self.update_setting_from_qt_view)
        self.radioButton_copy.clicked.connect(self.update_setting_from_qt_view)
        self.radioButton_sleep.clicked.connect(self.update_setting_from_qt_view)
        self.radioButton_setProxy.clicked.connect(self.update_setting_from_qt_view)
        self.radioButton_unlink_all_variable.clicked.connect(self.update_setting_from_qt_view)
        self.spinBox.valueChanged.connect(self.update_setting_from_qt_view)
        self.checkBox_autosend.clicked.connect(self.update_setting_from_qt_view)


        self.pushButton.clicked.connect(self.run)

    def update_parameters(self):
        self.str_proxy = (self.edit_proxy.text() or "").strip()

    def run(self):
        if self.thread is not None:
            self.thread.safe_quit()
        if self.in_data is None:
            self.Outputs.data.send(None)
            return

        self.error("")
        if self.str_radio_box_checked == "6":
            self.Outputs.data.send(unlink_domain(self.in_data))
            return

        # time peut importe l entree
        if self.str_radio_box_checked == "4":
            self.progressBarInit()
            self.thread = thread_management.Thread(OperationSystem.sleep_seconds, int(self.str_spinBox))
            self.thread.progress.connect(self.handle_progress)
            self.thread.result.connect(self.handle_result)
            self.thread.finish.connect(self.handle_finish)
            self.thread.start()
            return
        self.error("")
        if self.str_radio_box_checked == "5":
            self.progressBarInit()
            self.thread = thread_management.Thread(OperationSystem.set_proxy,self.str_proxy)
            self.thread.progress.connect(self.handle_progress)
            self.thread.result.connect(self.handle_result)
            self.thread.finish.connect(self.handle_finish)
            self.thread.start()
            return
        self.error("")
        try:
            self.in_data.domain["input"]
        except KeyError:
            self.error('You need a "input" column in input data')
            self.Outputs.data.send(None)
            return

        if type(self.in_data.domain["input"]).__name__ != 'StringVariable':
            self.error('"input" column needs to be a Text')
            return
        # liste en ignorant les valeurs manquante
        input_paths = [str(x) for x in self.in_data.get_column("input") if x is not None and str(x).strip() != ""]
        if len(input_paths)==0:
            self.error('"input" needs paths')
            return
        if self.str_radio_box_checked == "0":
            self.progressBarInit()
            self.thread = thread_management.Thread(OperationSystem.DeleteFolderlistOrFilelist, input_paths)
            self.thread.progress.connect(self.handle_progress)
            self.thread.result.connect(self.handle_result)
            self.thread.finish.connect(self.handle_finish)
            self.thread.start()
            return
        if self.str_radio_box_checked == "1":
            self.progressBarInit()
            self.thread = thread_management.Thread(OperationSystem.ensure_dirs, input_paths)
            self.thread.progress.connect(self.handle_progress)
            self.thread.result.connect(self.handle_result)
            self.thread.finish.connect(self.handle_finish)
            self.thread.start()
            return

        try:
            self.in_data.domain["output"]
        except KeyError:
            self.error('You need a "output" column in input data')
            return

        if type(self.in_data.domain["output"]).__name__ != 'StringVariable':
            self.error('"output" column needs to be a Text')
            return
        outputs_paths = [str(x) for x in self.in_data.get_column("output") if x is not None and str(x).strip() != ""]
        if len(input_paths)!=len(outputs_paths):
            self.error('"input" and "outputs" needs same dimensions')
            return
        if self.str_radio_box_checked == "2":
            self.progressBarInit()
            self.thread = thread_management.Thread(OperationSystem.move_or_rename, input_paths,outputs_paths)
            self.thread.progress.connect(self.handle_progress)
            self.thread.result.connect(self.handle_result)
            self.thread.finish.connect(self.handle_finish)
            self.thread.start()
            return
        if self.str_radio_box_checked == "3":
            self.progressBarInit()
            self.thread = thread_management.Thread(OperationSystem.copy_and_overwrite, input_paths,outputs_paths)
            self.thread.progress.connect(self.handle_progress)
            self.thread.result.connect(self.handle_result)
            self.thread.finish.connect(self.handle_finish)
            self.thread.start()
            return





    def handle_progress(self, value: float) -> None:
        self.progressBarSet(value)

    def handle_result(self, result):
        try:
            self.Outputs.data.send(self.in_data)
        except Exception as e:
            print("An error occurred when sending out_data:", e)
            self.Outputs.data.send(None)
            return
    def handle_finish(self):
        print("process finished")
        self.progressBarFinished()

# Test standalone
if __name__ == "__main__":
    import sys
    app = QApplication(sys.argv)
    window = OwOperationSystem()
    window.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()


import numpy as np
import re

import os
from AnyQt.QtCore import (
    Qt, QLineF, QSize, QAbstractTableModel, QModelIndex, QSortFilterProxyModel
)
from AnyQt.QtGui import QPainter, QPen, QBrush, QColor
from AnyQt.QtWidgets import (
    QItemDelegate, QHeaderView, QApplication
)

from Orange.classification.rules import _RuleClassifier
from Orange.widgets import widget, gui, settings
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.widget import Input, Output
from Orange.data import Table, Domain, ContinuousVariable, StringVariable



class OWRuleViewer(widget.OWWidget):
    name = "CN2 Rule Selector"
    description = "Review rules induced from data."
    icon = "icons/CN2RuleViewer.svg"
    category = "AAIT - ALGORITHM"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/CN2RuleViewer.svg"
    priority = 1140
    keywords = "cn2 rule viewer"
    #category="AAIT Resources Manager"
    class Inputs:
        #data = Input("Data", Table)
        point_from_scatter_plot= Input("Point From Scatter Plot", Table)
        classifier = Input("Classifier", _RuleClassifier)

    class Outputs:
#        annotated_data=Output("Annotated Data", Table)
#        selected_data = Output("Selected Data", Table, default=True)
        selected_rule =Output("Selected Rule", Table)


    compact_view = settings.Setting(False)

    want_basic_layout = True
    want_main_area = False
    want_control_area = True

    def __init__(self):
        super().__init__()

        self.data = None
        self.classifier = None
        self.selected = None
        self.data_point_from_scatter_plot =None
        self.model = CustomRuleViewerTableModel(parent=self)
        self.model.set_horizontal_header_labels(
            ["IF conditions", "", "THEN class", "Distribution",
             "Probabilities [%]", "Quality", "Length","number of tested data in rules"])

        self.proxy_model = QSortFilterProxyModel(parent=self)
        self.proxy_model.setSourceModel(self.model)
        self.proxy_model.setSortRole(self.model.SortRole)

        self.view = gui.TableView(self, wordWrap=False)
        self.view.setModel(self.proxy_model)
        self.view.verticalHeader().setVisible(True)
        self.view.horizontalHeader().setStretchLastSection(False)
        self.view.selectionModel().selectionChanged.connect(self.commit)

        self.dist_item_delegate = DistributionItemDelegate(self)
        self.view.setItemDelegateForColumn(3, self.dist_item_delegate)

        self.controlArea.layout().addWidget(self.view)

        gui.checkBox(widget=self.buttonsArea, master=self, value="compact_view",
                     label="Compact view", callback=self.on_update)
        gui.rubber(self.buttonsArea)

        original_order_button = gui.button(
            self.buttonsArea, self,
            "Restore original order",
            autoDefault=False,
            callback=self.restore_original_order,
            attribute=Qt.WA_LayoutUsesWidgetRect,
        )
        original_order_button.clicked.connect(self.restore_original_order)

    # @Inputs.data
    # def set_data(self, data):
    #     self.data = data
    #     self.commit()

    @Inputs.classifier
    def set_classifier(self, classifier):
        self.classifier = classifier
        self.selected = None
        self.model.clear()

        if classifier is not None and hasattr(classifier, "rule_list"):
            self.model.set_vertical_header_labels(
                list(range(len(classifier.rule_list))))

            self.dist_item_delegate.color_schema = \
                [QColor(*c) for c in classifier.domain.class_var.colors]

            self.model.wrap(self.classifier.domain, self.classifier.rule_list)

        self.on_update()
        self.commit()

    @Inputs.point_from_scatter_plot
    def set_new_point_from_scatter_plot(self, data):
        self.data_point_from_scatter_plot = data
        self.on_update()
    def on_update(self):
        self.model.wrap_data_point_from_scatter_plot(self.data_point_from_scatter_plot)
        self._save_selected()

        self.model.set_compact_view(self.compact_view)
        if self.compact_view:
            self.view.horizontalHeader().setSectionResizeMode(
                0, QHeaderView.Interactive)  # QHeaderView.Stretch
        else:
            self.view.horizontalHeader().setSectionResizeMode(
                QHeaderView.ResizeToContents)
        self.view.resizeColumnsToContents()
        self.view.resizeRowsToContents()

        self._restore_selected()

    def _save_selected(self, actual=False):
        self.selected = None
        selection_model = self.view.selectionModel()
        if selection_model.hasSelection():
            if not actual:
                selection = selection_model.selection()
            else:
                selection = self.proxy_model.mapSelectionToSource(
                    selection_model.selection())

            self.selected = sorted(set(index.row() for index
                                       in selection.indexes()))

    def _restore_selected(self):
        if self.selected is not None:
            selection_model = self.view.selectionModel()
            for row in self.selected:
                selection_model.select(self.proxy_model.index(row, 0),
                                       selection_model.Select |
                                       selection_model.Rows)

    def restore_original_order(self):
        self.proxy_model.sort(-1)

    def copy_to_clipboard(self):
        self._save_selected(actual=True)
        if self.selected is not None:
            output = "\n".join([str(self.classifier.rule_list[i])
                                for i in self.selected])
            QApplication.clipboard().setText(output)

    def commit(self):

        self._save_selected(actual=True)
        # selected_indices = []

        data = self.data or self.classifier and self.classifier.instances
        if (self.selected is not None and
                data is not None and
                self.classifier is not None and
                data.domain.attributes ==
                self.classifier.original_domain.attributes):

            status = np.ones(data.X.shape[0], dtype=bool)
            for i in self.selected:
                rule = self.classifier.rule_list[i]
                status &= rule.evaluate_data(data.X)

            selected_indices = status.nonzero()[0]
            if len(selected_indices)>0:
                if len(status) != len(data):
                    raise ValueError("La longueur de status doit correspondre au nombre de lignes dans data.")

                # status_var = StringVariable("Status")
                # new_domain = Domain(data.domain.attributes, data.domain.class_vars,
                #                     metas=data.domain.metas + (status_var,))
                # outdata_status = Table.from_table(new_domain, data)
                # tab_status=status.reshape(-1, 1)
                # outdata_status[:, "Status"] = tab_status.astype(str)
                # self.Outputs.annotated_data.send(outdata_status)
                # Déclarer la nouvelle variable continue comme attribut
                status_var = ContinuousVariable("Status")
                new_domain = Domain(data.domain.attributes + (status_var,), data.domain.class_vars, data.domain.metas)

                # Créer une nouvelle table basée sur le domaine mis à jour
                outdata_status = Table.from_table(new_domain, data)

                # Convertir "True"/"False" en valeurs continues (1 ou 100)
                tab_status = (status == True).astype(int) * 99 + 1  # True -> 100, False -> 1
                tab_status = tab_status.reshape(-1, 1)

                # Ajouter les valeurs continues à la nouvelle variable
                outdata_status[:, "Status"] = tab_status

                # Envoyer les données modifiées
#                self.Outputs.annotated_data.send(outdata_status)

            # a verifier surement fausse alarme pyflakes
            # data_output = data.from_table_rows(data, selected_indices) \
            #     if len(selected_indices) else None
        if self.classifier!=None:
            if self.selected!=[] and self.selected is not None:
                data = []
                meta = []

                for index_select in self.selected:
                    rule=(self.classifier.rule_list[index_select])
                    regle = str(rule).split(" THEN ")[0]
                    proba = 0
                    if rule.prediction == 0:
                        proba = rule.probabilities[0]
                    else:
                        proba = rule.probabilities[1]
                    data.append([rule.prediction, proba])
                    new_regle=regle
                    if len(new_regle)>3:
                        if new_regle[:3]=="IF ":
                            new_regle=new_regle[3:]
                    # meta.append(new_regle)
                    new_regle=new_regle.replace(" AND ", " and ")
                    #meta.append([regle.replace("IF", "").replace("THEN", "").replace("_", "").replace(" AND ", " and ")])
                    meta.append([new_regle])
                # header = ['prediction', 'quality']
                domain = Domain([ContinuousVariable('prediction'), ContinuousVariable('proba')],
                                metas=[StringVariable('regle')])

                # domain = Domain([ContinuousVariable('prediction'), ContinuousVariable('proba'), DiscreteVariable('regle', values=meta)] )
                out_data = Table.from_numpy(domain, data, metas=meta)
                self.Outputs.selected_rule.send(out_data)
#        self.Outputs.selected_data.send(data_output)


    def send_report(self):
        if self.classifier is not None:
            self.report_table("Induced rules", self.view)

    def sizeHint(self):
        return QSize(800, 450)


class CustomRuleViewerTableModel(QAbstractTableModel):
    SortRole = Qt.UserRole + 1

    # ascii operators
    OPERATORS = {
        '==': '=',
        '!=': '≠',
        '<=': '≤',
        '>=': '≥'
    }

    def __init__(self, parent=None):
        super().__init__(parent)
        self._domain = None
        self._rule_list = []
        self._compact_view = False
        self._headers = {}
        self._data_point_from_scatter_plot=None

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            headers = self._headers.get(orientation)
            return (headers[section] if headers and section < len(headers)
                    else str(section))
        return None

    def del_space_debut_fin(self, text_to_edit):
        if text_to_edit[0] == " ":
            text_to_edit = text_to_edit[1:]
        if text_to_edit[-1] == " ":
            text_to_edit = text_to_edit[:-1]
        return text_to_edit

    def set_horizontal_header_labels(self, labels):
        self._headers[Qt.Horizontal] = labels

    def set_vertical_header_labels(self, labels):
        self._headers[Qt.Vertical] = labels

    def rowCount(self, parent=QModelIndex()):
        return 0 if parent.isValid() else len(self)

    def columnCount(self, parent=QModelIndex()):
        return 0 if parent.isValid() else len(self._headers[Qt.Horizontal])

    def wrap(self, domain, rule_list):
        self.beginResetModel()
        self._domain = domain
        self._rule_list = rule_list
        self.endResetModel()
    def wrap_data_point_from_scatter_plot(self,data_point_from_scatter_plot):
        self._data_point_from_scatter_plot=data_point_from_scatter_plot
    def clear(self):
        self.beginResetModel()
        self._domain = None
        self._rule_list = []
        self.endResetModel()

    def set_compact_view(self, compact_view):
        self.beginResetModel()
        self._compact_view = compact_view
        self.endResetModel()

    def data(self, index, role=Qt.DisplayRole):
        if self._domain is None or not index.isValid():
            return None

        def _display_role():
            if column == 0:
                delim = " AND " if self._compact_view else " AND\n"
                return "TRUE" if not rule.selectors else delim.join(
                    [attributes[s.column].name + self.OPERATORS[s.op] +
                     (attributes[s.column].values[int(s.value)]
                      if attributes[s.column].is_discrete
                      else str(s.value)) for s in rule.selectors])
            if column == 1:
                return '→'
            if column == 2:
                return class_var.name + "=" + class_var.values[rule.prediction]
            if column == 3:
                # type(curr_class_dist) = ndarray
                return ([float(format(x, '.1f')) for x in rule.curr_class_dist]
                        if rule.curr_class_dist.dtype == float
                        else rule.curr_class_dist.tolist())
            if column == 4:
                return " : ".join(str(int(round(100 * x)))
                                  for x in rule.probabilities)
            if column == 5:
                value = rule.quality
                absval = abs(value)
                strlen = len(str(int(absval)))
                return '{:.{}{}}'.format(value,
                                         2 if absval < .001 else
                                         3 if strlen < 2 else
                                         1 if strlen < 5 else
                                         0 if strlen < 6 else
                                         3,
                                         'f' if (absval == 0 or
                                                 absval >= .001 and
                                                 strlen < 6)
                                         else 'e')
            if column == 6:
                return rule.length
            if column == 7:
                if self._data_point_from_scatter_plot==None:
                    return 0
                if len(self._data_point_from_scatter_plot)==0:
                    return 0
                input_selected_data = self._data_point_from_scatter_plot
                line=str(rule)
                line =line[3:]# remove IF
                current_rules=line.split(" THEN ")[0]
                if current_rules == "TRUE":
                    return len(input_selected_data)
                regl_list = current_rules.split(" AND ")
                regle_a_tester = []
                for unit_rule in regl_list:
                    current_var, current_symb, current_value = re.split(r'(<=|>=)', unit_rule)
                    current_var = self.del_space_debut_fin(current_var)
                    current_symb = self.del_space_debut_fin(current_symb)
                    current_value = float(self.del_space_debut_fin(current_value))
                    regle_a_tester.append([current_var, current_symb, current_value])
                nb_dedans = 0
                for la_data in input_selected_data:
                    dedans = True
                    for regl in regle_a_tester:
                        if dedans == False:
                            break
                        if regl[1] == ">=":
                            if la_data[la_data.domain.index(regl[0])].value <= regl[2]:
                                dedans = False
                        else:
                            if la_data[la_data.domain.index(regl[0])].value >= regl[2]:
                                dedans = False
                    if dedans:
                        nb_dedans = nb_dedans + 1
                return nb_dedans



                return 0

            return None

        def _tooltip_role():
            if column == 0:
                return _display_role().replace(" AND ", " AND\n")
            if column == 1:
                return None
            if column == 3:
                # list of int, float
                curr_class_dist = _display_role()
                return class_var.name + "\n" + "\n".join(
                    (str(curr_class) + ": " + str(curr_class_dist[i])
                     for i, curr_class in enumerate(class_var.values)))
            if column == 4:
                return class_var.name + "\n" + "\n".join(
                    str(curr_class) + ": " +
                    '{:.1f}'.format(rule.probabilities[i] * 100) + "%"
                    for i, curr_class in enumerate(class_var.values))
            return _display_role()

        def _sort_role():
            if column == 0:
                return rule.length
            if column == 3:
                # return int, not np.int!
                return int(sum(rule.curr_class_dist))
            return _display_role()

        attributes = self._domain.attributes
        class_var = self._domain.class_var
        rule = self._rule_list[index.row()]
        column = index.column()

        if role == Qt.DisplayRole:
            return _display_role()
        if role == Qt.ToolTipRole:
            return _tooltip_role()
        if role == self.SortRole:
            return _sort_role()
        if role == Qt.TextAlignmentRole:
            return (Qt.AlignVCenter | [(Qt.AlignRight if self._compact_view
                                        else Qt.AlignLeft), Qt.AlignCenter,
                                       Qt.AlignLeft, Qt.AlignCenter,
                                       Qt.AlignCenter, Qt.AlignRight,
                                       Qt.AlignRight,Qt.AlignRight][column])

    def __len__(self):
        return len(self._rule_list)

    def __bool__(self):
        return len(self) != 0

    def __iter__(self):
        return iter(self._rule_list)

    def __getitem__(self, item):
        return self._rule_list[item]


class DistributionItemDelegate(QItemDelegate):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.color_schema = None

    def sizeHint(self, option, index):
        size = super().sizeHint(option, index)
        size.setWidth(size.width() + 10)
        return size

    def paint(self, painter, option, index):
        curr_class_dist = np.array(index.data(Qt.DisplayRole), dtype=float)
        curr_class_dist /= sum(curr_class_dist)
        painter.save()
        self.drawBackground(painter, option, index)
        rect = option.rect

        if sum(curr_class_dist) > 0:
            pw = 3
            hmargin = 5
            x = rect.left() + hmargin
            width = rect.width() - 2 * hmargin
            vmargin = 1
            textoffset = pw + vmargin * 2
            painter.save()
            baseline = rect.bottom() - textoffset / 2

            text = str(index.data(Qt.DisplayRole))
            option.displayAlignment = Qt.AlignCenter
            text_rect = rect.adjusted(0, 0, 0, -textoffset*0)
            self.drawDisplay(painter, option, text_rect, text)

            painter.setRenderHint(QPainter.Antialiasing)
            for prop, color in zip(curr_class_dist, self.color_schema):
                if prop == 0:
                    continue
                painter.setPen(QPen(QBrush(color), pw))
                to_x = x + prop * width
                line = QLineF(x, baseline, to_x, baseline)
                painter.drawLine(line)
                x = to_x
            painter.restore()

        painter.restore()


if __name__ == "__main__":  # pragma: no cover
    from Orange.classification import CN2Learner
    data = Table("iris")
    learner = CN2Learner()
    model = learner(data)
    model.instances = data
    WidgetPreview(OWRuleViewer).run(model)



import os
import sys
from pathlib import Path


def replace_owcorpus_file():
    if os.name == "nt":
        python_path = sys.executable.replace("\\","/")
        dir_path = os.path.dirname(python_path)
        owcorpus_file_path =dir_path+"/lib/site-packages/orangecontrib/text/widgets/owcorpus.py"
        file_py = Path(owcorpus_file_path)
        file_txt = Path(os.path.dirname(os.path.abspath(__file__)).replace("\\", "/") + "/owcorpus_ok.txt")
        if os.path.exists(dir_path+"/lib/site-packages/orangecontrib/text/widgets")==False:
            return
        size = os.stat(owcorpus_file_path).st_size
        if size != 14559:
            return
        os.remove(owcorpus_file_path)
        contenu = file_txt.read_text()
        file_py.write_text(contenu)
        print(f"{owcorpus_file_path} a été remplacé.")



if __name__ == "__main__":
    replace_owcorpus_file()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
orange_tab_viewer.py
--------------------
Viewer autonome (AnyQt) pour fichiers .tab (Orange Data Mining) :
- Lit un .tab via Orange.data.Table.from_file()
- Convertit toutes les valeurs en chaînes pour l'affichage
- Affiche une table moderne (thème sombre, tri par colonne)
- Colorise les cellules si valeur = "[color]texte" (ex. [red]Erreur)
- Optionnel : auto-refresh sur modification du fichier --watch <ms>

Dépendances : AnyQt, Orange3
"""

from __future__ import annotations
import sys, os, argparse, re
from typing import List, Tuple, Optional, Dict

from AnyQt.QtCore import Qt, QTimer, QFileSystemWatcher, QDateTime, QSize
from AnyQt.QtGui import QPalette, QColor, QFont, QStandardItemModel, QStandardItem, QAction
from AnyQt.QtWidgets import (
    QApplication, QMainWindow, QWidget, QVBoxLayout, QHBoxLayout, QLabel,
    QFileDialog, QTableView, QStatusBar, QToolBar, QAbstractItemView, QStyleFactory, QHeaderView
)

from Orange.data import Table, Domain

# ---------- Colorisation via tag "[color]texte" ----------
COLOR_TAG_RE = re.compile(r"^\s*\[(?P<color>[a-zA-Z]+)\]\s*(?P<text>.*)\s*$")

NAMED_COLORS: Dict[str, QColor] = {
    "red": QColor("#d84a4a"),
    "green": QColor("#2da44e"),
    "blue": QColor("#2f81f7"),
    "orange": QColor("#f29900"),
    "yellow": QColor("#e3b341"),
    "purple": QColor("#8957e5"),
    "magenta": QColor("#d527b7"),
    "pink": QColor("#ff5ca7"),
    "teal": QColor("#1f8a70"),
    "cyan": QColor("#00acc1"),
    "grey": QColor("#5b6470"),
    "gray": QColor("#5b6470"),
    "black": QColor("#111111"),
    "white": QColor("#ffffff"),
}

def best_foreground_for(bg: QColor) -> QColor:
    r, g, b = bg.red(), bg.green(), bg.blue()
    yiq = ((r*299) + (g*587) + (b*114)) / 1000
    return QColor("#000000") if yiq > 150 else QColor("#ffffff")

# ---------- Palette sombre ----------
def nice_dark_palette() -> QPalette:
    p = QPalette()
    bg = QColor(11, 11, 15)
    panel = QColor(17, 23, 31)
    text = QColor(234, 234, 234)
    muted = QColor(170, 184, 200)
    highlight = QColor(30, 96, 180)
    p.setColor(QPalette.Window, bg)
    p.setColor(QPalette.Base, QColor(15, 20, 27))
    p.setColor(QPalette.AlternateBase, QColor(19, 26, 34))
    p.setColor(QPalette.Text, text)
    p.setColor(QPalette.Button, panel)
    p.setColor(QPalette.ButtonText, text)
    p.setColor(QPalette.Highlight, highlight)
    p.setColor(QPalette.HighlightedText, QColor(255, 255, 255))
    p.setColor(QPalette.ToolTipBase, panel)
    p.setColor(QPalette.ToolTipText, text)
    p.setColor(QPalette.WindowText, text)
    p.setColor(QPalette.PlaceholderText, muted)
    return p

# ---------- Conversion Orange Table -> matrice de chaînes ----------
def orange_table_to_strings(data: Table) -> Tuple[List[str], List[List[str]]]:
    """
    Convertit une Orange.data.Table en :
      - headers: liste de noms de colonnes
      - rows: liste de lignes (valeurs converties en str)
    Concatène attributes + class_vars + metas (dans cet ordre).
    """
    dom: Domain = data.domain
    attrs = list(dom.attributes)
    class_vars = list(dom.class_vars) if dom.class_vars else []
    metas = list(dom.metas) if dom.metas else []

    headers = [v.name for v in attrs] + [v.name for v in class_vars] + [str(v.name) for v in metas]

    cols = []
    for var in attrs + class_vars:
        col, _ = data.get_column_view(var)
        cols.append(col)
    for mv in metas:
        col, _ = data.get_column_view(mv)
        cols.append(col)

    n_rows = len(data)
    n_cols = len(cols)

    rows: List[List[str]] = []
    for i in range(n_rows):
        out_row: List[str] = []
        for j in range(n_cols):
            val = cols[j][i]
            s = "" if val is None else str(val)
            if s.lower() == "nan":
                s = ""
            out_row.append(s)
        rows.append(out_row)
    return headers, rows

# ---------- Modèle Qt ----------
class StringTableModel(QStandardItemModel):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.enable_color_tags: bool = True

    def load(self, headers: List[str], rows: List[List[str]]):
        self.clear()
        self.setColumnCount(len(headers))
        self.setHorizontalHeaderLabels(headers)
        self.setRowCount(len(rows))

        for r, row in enumerate(rows):
            for c, raw_value in enumerate(row):
                display_text, bg_color = self._parse_color_tag(raw_value) if self.enable_color_tags else (raw_value, None)
                item = QStandardItem(display_text)
                item.setEditable(False)
                if bg_color is not None:
                    fg = best_foreground_for(bg_color)
                    item.setBackground(bg_color)
                    item.setForeground(fg)
                self.setItem(r, c, item)

    @staticmethod
    def _parse_color_tag(value: str):
        if not isinstance(value, str):
            return str(value), None
        m = COLOR_TAG_RE.match(value)
        if not m:
            return value, None
        color_name = m.group("color").strip().lower()
        text = m.group("text")
        qcolor = NAMED_COLORS.get(color_name)
        return text, qcolor

# ---------- Fenêtre principale ----------
class MainWindow(QMainWindow):
    def __init__(self, path: Optional[str], watch_ms: Optional[int]):
        super().__init__()
        self.setWindowTitle("Orange .tab Viewer — AnyQt")
        self.setMinimumSize(QSize(1000, 600))

        self.path: Optional[str] = path
        self.watch_ms: Optional[int] = watch_ms
        self.last_mtime: Optional[float] = None

        self._apply_style()
        self._setup_ui()
        self._setup_toolbar()
        self._setup_watch()

        if self.path:
            self.load_tab(self.path)

    def _apply_style(self):
        QApplication.setStyle(QStyleFactory.create("Fusion"))
        QApplication.setPalette(nice_dark_palette())
        f = QFont(); f.setPointSize(10)
        QApplication.setFont(f)

    def _setup_ui(self):
        central = QWidget(self); self.setCentralWidget(central)
        lay = QVBoxLayout(central); lay.setContentsMargins(12, 10, 12, 10); lay.setSpacing(8)

        top = QHBoxLayout()
        self.info = QLabel("Aucun fichier chargé")
        self.info.setStyleSheet("color:#9ad0ff;")
        self.color_hint = QLabel("Tags couleur : [red]/[green]/[blue]/[orange]/…")
        self.color_hint.setStyleSheet("color:#8aa0b2;")
        top.addWidget(self.info, 1); top.addWidget(self.color_hint, 0)
        lay.addLayout(top)

        self.table = QTableView()
        self.table.setAlternatingRowColors(True)
        self.table.setSortingEnabled(True)
        self.table.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.table.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self.table.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeMode.Interactive)
        self.table.horizontalHeader().setStretchLastSection(True)
        self.table.verticalHeader().setVisible(False)
        self.table.setCornerButtonEnabled(False)
        self.table.setShowGrid(False)
        self.table.setStyleSheet("""
            QTableView {
                gridline-color: #1f2630;
                selection-background-color: #1e60b4;
                selection-color: #ffffff;
                outline: 0;
                border: 1px solid #2a3442;
                border-radius: 10px;
                background: #0f141b;
            }
            QHeaderView::section {
                background: #111827;
                color: #9ad0ff;
                border: 0;
                padding: 8px;
            }
        """)
        self.model = StringTableModel(self)
        self.table.setModel(self.model)
        lay.addWidget(self.table, 1)

        sb = QStatusBar(self)
        sb.setStyleSheet("QStatusBar{border-top:1px solid #1f2630;}")
        self.setStatusBar(sb)

    def _setup_toolbar(self):
        tb = QToolBar("Fichier", self)
        tb.setIconSize(QSize(18, 18))
        self.addToolBar(Qt.ToolBarArea.TopToolBarArea, tb)

        act_open = QAction("Ouvrir…", self)
        act_open.triggered.connect(self.choose_file)
        tb.addAction(act_open)

        act_reload = QAction("Recharger", self)
        act_reload.triggered.connect(self.reload)
        tb.addAction(act_reload)

        tb.addSeparator()

        act_quit = QAction("Quitter", self)
        act_quit.triggered.connect(self.close)
        tb.addAction(act_quit)

    def _setup_watch(self):
        self.watcher = QFileSystemWatcher(self)
        self.watcher.fileChanged.connect(self._on_fs_change)
        self.watcher.directoryChanged.connect(self._on_fs_change)

        self.poll = QTimer(self)
        self.poll.timeout.connect(self._check_mtime)
        if self.watch_ms and self.watch_ms > 0:
            self.poll.start(max(300, self.watch_ms))

    # ---- Fichier .tab ----
    def choose_file(self):
        path, _ = QFileDialog.getOpenFileName(self, "Ouvrir un .tab Orange", os.path.dirname(self.path or ".") or ".", "Orange .tab (*.tab);;Tous les fichiers (*.*)")
        if not path:
            return
        self.load_tab(path)

    def load_tab(self, path: str):
        try:
            data = Table.from_file(path)
        except Exception as e:
            self.statusBar().showMessage(f"Erreur de lecture: {e}", 6000)
            return

        headers, rows = orange_table_to_strings(data)
        self.model.load(headers, rows)
        self.table.resizeColumnsToContents()
        for c in range(self.model.columnCount()):
            w = max(self.table.columnWidth(c), 80)
            self.table.setColumnWidth(c, min(w, 420))

        self.path = path
        self.last_mtime = self._safe_mtime(path)
        self._watch_register(path)

        ts = QDateTime.currentDateTime().toString("yyyy-MM-dd HH:mm:ss")
        self.info.setText(f"{os.path.basename(path)} — {len(rows)} lignes, {len(headers)} colonnes  |  MAJ {ts}")
        self.statusBar().showMessage("Chargé.", 2000)

    def reload(self):
        if not self.path:
            return
        self.load_tab(self.path)

    # ---- Watch helpers ----
    def _watch_register(self, path: str):
        try:
            self.watcher.removePaths(self.watcher.files() + self.watcher.directories())
        except Exception:
            pass
        if os.path.isfile(path):
            self.watcher.addPath(path)
        folder = os.path.dirname(path) or "."
        if os.path.isdir(folder):
            self.watcher.addPath(folder)

    def _on_fs_change(self, _):
        # petit délai pour laisser l'éditeur terminer l'écriture
        QTimer.singleShot(150, self.reload)

    def _safe_mtime(self, path: str) -> Optional[float]:
        try:
            return os.path.getmtime(path)
        except Exception:
            return None

    def _check_mtime(self):
        if not self.path:
            return
        new_m = self._safe_mtime(self.path)
        if new_m is None:
            return
        if self.last_mtime is None or new_m != self.last_mtime:
            self.last_mtime = new_m
            self.reload()

# ---------- CLI ----------
def parse_args(argv: list[str]) -> argparse.Namespace:
    ap = argparse.ArgumentParser(description="Viewer .tab Orange (casting string + color tags).")
    ap.add_argument("tab", nargs="?", help="Chemin du fichier .tab")
    ap.add_argument("--watch", type=int, default=0, help="Intervalle de vérification (ms). 0 = désactivé.")
    return ap.parse_args(argv)

def main():
    args = parse_args(sys.argv[1:])
    app = QApplication(sys.argv)
    win = MainWindow(path=args.tab, watch_ms=args.watch)
    win.show()
    return app.exec()

if __name__ == "__main__":
    raise SystemExit(main())



import json
import os
import subprocess
import zipfile
# import requests
from pathlib import Path
from bs4 import BeautifulSoup
from urllib.parse import urljoin
if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils import SimpleDialogQt
else:
    from orangecontrib.AAIT.utils import SimpleDialogQt
try:
    from ..utils import MetManagement
except:
    from Orange.widgets.orangecontrib.AAIT.utils import MetManagement



def generate_listing_json(directory_to_list):
    """
    generate a file file_info.json  with size of file in directory
    for low level devellopper only
    -> use create_index_file
    """
    output_json_file = directory_to_list+'/files_info.json'
    if os.path.isfile(output_json_file):
        os.remove(output_json_file)

    def list_files_recursive(directory):
        files_info = {}

        for root, dirs, files in os.walk(directory):
            for file in files:

                file_path = os.path.join(root, file).replace("\\","/")
                file_size = os.path.getsize(file_path)
                file_path=file_path[len(directory)+1:]
                files_info[file_path]=file_size

        return files_info

    def save_to_json(data, output_file):
        with open(output_file, 'w', encoding='utf-8') as f:
            json.dump(data, f, ensure_ascii=False, indent=4)

    files_info = list_files_recursive(directory_to_list)
    save_to_json(files_info, output_json_file)
    return files_info

def add_file_to_zip_file(folder,file_in,zip_file):
    """
    folder : a point to start relative path
    file in : a file to add to zip file
    zip file : destination zip file
    for exemple I want to add C:/dir1/dir2/dir3/qwerty.txt to
    C:/dir1//dir2/example.zip and index dir3/qwerty.txt
    folder = C:/dir1/
    file_in=C:/dir1/dir2/dir3/qwerty.txt
    zip_file  C:/dir2/example.zip
    """
    path_in=folder+file_in
    with open(path_in, 'rb') as f:
        contenu = f.read()
    with zipfile.ZipFile(zip_file, 'a', zipfile.ZIP_DEFLATED) as archive:
        archive.writestr(file_in, contenu)

def create_index_file(in_repo_file,out_repo_file="a_ignorer"):
    """
    delete files_info.json and regenerate it
    with file contain a dictionnary filename:filesize
    out_repo_file is not used
    """
    if not os.path.isfile(in_repo_file):
        raise Exception("The specified path is not a file "+in_repo_file)
    if 0!=MetManagement.get_size(in_repo_file):
        raise Exception("The file " + in_repo_file+ " need to be empty to use this functions")
    print(MetManagement.get_size(in_repo_file))
    folder_to_process=os.path.dirname(in_repo_file).replace("\\","/")
    file_info=generate_listing_json(folder_to_process)
    print(file_info)
    return

def decode(repo_file,file_to_read):
    """
    be carrefull with big file (ram saturation)
    return containt of a zipped file
    """
    if not os.path.isfile(repo_file):
        return None
    file_to_read=os.path.splitext(os.path.basename(repo_file))[0]+"/"+file_to_read
    with zipfile.ZipFile(repo_file, 'r') as zip_ref:
        with zip_ref.open(file_to_read) as file:
            content = file.read()
            return content.decode('utf-8')
def decode_to_file(zip_path, target_path, output_path):
    """
    extract a file from a zip file and write it on hdd
    example : I want to extract dir1/qwerty.txt from C:/dir1/dir2/zipfile.zip to C:/dir_a/dir_b/dir1/qwerty.txt
    zip_path=C:/dir1/dir2/zipfile.zip
    target_path=dir1/qwerty.txt
    output_path=C:/dir_a/dir_b/
    """
    chunk_size = 1024 * 1024 * 100 # 100 Mo
    target_path=os.path.splitext(os.path.basename(zip_path))[0]+"/"+target_path
    with zipfile.ZipFile(zip_path, 'r') as zip_ref:
        target_path = target_path.rstrip('/')
        files_to_extract = [f for f in zip_ref.namelist() if f.startswith(target_path)]

        if len(files_to_extract) == 0:
            raise FileNotFoundError(f"{target_path} not found in the archive.")

        if len(files_to_extract) == 1 and not files_to_extract[0].endswith('/'):
            # Cible est un fichier unique
            output_file_path = output_path
            os.makedirs(os.path.dirname(output_file_path), exist_ok=True)
            with zip_ref.open(files_to_extract[0]) as source, open(output_file_path, 'wb') as target:
                #target.write(source.read())
                while True:
                    # read and write a chunk to avoid ram limitation
                    chunk = source.read(chunk_size)
                    if not chunk:
                        break
                    target.write(chunk)
        else:
            # Cible est un dossier ou plusieurs fichiers
            for file in files_to_extract:
                relative_path = os.path.relpath(file, start=target_path)
                destination_path = os.path.join(output_path, relative_path)

                if file.endswith('/'):
                    os.makedirs(destination_path, exist_ok=True)
                else:
                    os.makedirs(os.path.dirname(destination_path), exist_ok=True)
                    with zip_ref.open(file) as source, open(destination_path, 'wb') as target:
                        #target.write(source.read())
                        while True:
                            # read and write a chunk to avoid ram limitation
                            chunk = source.read(chunk_size)
                            if not chunk:
                                break
                            target.write(chunk)

def normalize_path(path):
    """
    Normalize paths for URLs and local usage:
    - Replaces backslashes with forward slashes .
    - Removes './' and '\\.' segments from paths.
    - Handles redundant slashes.
    """
    # Replace backslashes with slashes
    path = path.replace("\\", "/")

    # Remove any occurrences of './' or '\.'
    path = path.replace("./", "").replace("/./", "/")

    # Clean up multiple slashes (e.g., "///" -> "/")
    path = os.path.normpath(path).replace("\\", "/")

    # Remove trailing and leading slashes (if required)
    path = path.strip("/")

    return path


def normalize_child_name(href):
    """
    Cleans and normalizes the given path or file name.

    This function takes a path or file name as input, and performs the following operations:
    1. Strips leading and trailing whitespace from the input.
    2. Normalizes the path using `os.path.normpath`, which:
        - Removes redundant separators (e.g., '///' -> '/').
        - Removes up-level references (e.g., '/../' -> '').
    3. Removes leading './' from the path if present.

    Parameters:
    href (str): The path or file name to normalize.

    Returns:
    str: The normalized path or file name.
    """
    # 1. Strip leading and trailing whitespace
    child = href.strip()

    # 2. Normalize the path
    child = os.path.normpath(child)

    # 3. Remove leading "./" if present
    if child.startswith("." + os.sep):
        child = child[len("." + os.sep):]

    return child


def download_from_folder_server(base_url: str, local_dir: str = ".", target_subfolder: str = "",
                                visited: set = None) -> None:
    """
    Recursively retrieves all files and subdirectories from the subfolder
    `target_subfolder` on the remote server (base_url) and recreates the structure in the local directory (local_dir).

    :param base_url: Base URL of the server (e.g., "http://server.com/folder").
    :param local_dir: Local path where the downloaded files will be stored.
    :param target_subfolder: Relative path (on the server) to download.
    :param visited: Set of URLs already visited to avoid loops.
    """
    if visited is None:
        visited = set()

    # Build the full URL from the base and the relative path.
    full_url = urljoin(base_url.rstrip('/') + '/', target_subfolder.strip('/'))
    if full_url in visited:
        return
    visited.add(full_url)

    # Determine the local destination path based on target_subfolder.
    # If target_subfolder ends with an extension, it is considered a file.
    local_target_dir = Path(local_dir) / target_subfolder.strip('/')
    is_file = os.path.splitext(target_subfolder)[1] != ""

    if is_file:
        # Direct file download.
        local_filename = local_target_dir
        dialog_progress = SimpleDialogQt.ProgressDialog(
            title="Synchronization, please wait",
            content=("Synchronization of the file " + full_url + "\nThis operation may take a few minutes"))
        dialog_progress.show()
        if MetManagement.already_downloaded_compressed_server(base_url, target_subfolder,
                                                              str(local_filename.resolve())):
            print(f"{local_filename.resolve()} already downloaded -> skip")
            return

        local_filename.parent.mkdir(parents=True, exist_ok=True)
        curl_command = f'curl -L -# -C - -o "{local_filename.resolve()}" "{full_url}"'
        try:
            creation_flags = subprocess.CREATE_NO_WINDOW if os.name == 'nt' else 0
            process = subprocess.Popen(
                curl_command,
                shell=True,
                text=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                creationflags=creation_flags
            )
            while True:
                line = process.stdout.readline()
                if line == "" and process.poll() is not None:
                    break
                if line:
                    print(line.strip())

            retcode = process.poll()
            if retcode != 0:
                print(f"Download failed for {full_url} with return code {retcode}")
            else:
                if not local_filename.exists():
                    print(f"Error: the file {local_filename} does not exist after download.")
                else:
                    print(f"File downloaded: {local_filename.resolve()}")
                    dialog_progress.stop()
        except Exception as e:
            print(f"Exception during download of {full_url}: {e}")
        return

    else:
        creation_flags = subprocess.CREATE_NO_WINDOW if os.name == 'nt' else 0

        # Case for a folder: retrieve the listing using curl.
        curl_command = f'curl -L -# -C - "{full_url}"'
        process = subprocess.Popen(curl_command, shell=True,text=True,
                                   stdout=subprocess.PIPE, stderr=subprocess.STDOUT,creationflags=creation_flags)
        html_lines = []
        while True:
            line = process.stdout.readline()
            print(type(line),line)
            if line == "" and process.poll() is not None:
                break
            if line:
                html_lines.append(line)
        retcode = process.poll()
        if retcode != 0:

            print(f"Failed to retrieve listing for {full_url} with code {retcode}")
            # Attempt direct download in this case:
            local_filename = local_target_dir
            local_filename.parent.mkdir(parents=True, exist_ok=True)
            # Build the curl command
            curl_command = f'curl -L -# -C - -o "{local_filename.resolve()}" "{full_url}"'
            dialog_progress = SimpleDialogQt.ProgressDialog(
                title="Synchronization, please wait",
                content=("Synchronization of the file " + full_url + "\nThis operation may take a few minutes"))
            dialog_progress.show()
            # On Windows, add the flag to avoid creating a new window.
            creation_flags = subprocess.CREATE_NO_WINDOW if os.name == 'nt' else 0

            # Run the curl command with no output and no window.
            process = subprocess.Popen(
                curl_command,
                shell=True,
                text=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                creationflags=creation_flags
            )
            if process.returncode != 0:
                print(f"Direct download failed for {full_url}: {process.stderr}")
            else:
                print(f"File downloaded: {local_filename.resolve()}")
                dialog_progress.stop()
            return

        html_content = "".join(html_lines)
        # Parse the HTML content to detect a directory listing.
        soup = BeautifulSoup(html_content, 'html.parser')
        h1 = soup.find('h1')
        has_listing_indicator = (h1 and "contenu de" in h1.get_text().lower())
        table = soup.find('table')

        if has_listing_indicator and table:
            rows = table.find_all('tr')
            for row in rows:
                links = row.find_all('a', href=True)
                if not links:
                    continue
                link = links[0]
                href = link['href']
                if href.startswith('../'):
                    continue

                # Normalize the name using the dedicated function.
                child_name = normalize_child_name(href)
                child_url = urljoin(full_url + '/', href)
                if href.endswith('/'):
                    # Subfolder detected: recursive call.
                    new_target_subfolder = os.path.join(target_subfolder.strip('/'), child_name)
                    new_target_subfolder = os.path.normpath(new_target_subfolder).replace('\\', '/')
                    download_from_folder_server(
                        base_url=base_url,
                        local_dir=local_dir,
                        target_subfolder=new_target_subfolder,
                        visited=visited
                    )
                else:
                    # File detected in the listing.
                    local_filename = local_target_dir / child_name
                    local_filename_parsed = str(local_filename).replace("\\", "/")
                    creation_flags = subprocess.CREATE_NO_WINDOW if os.name == 'nt' else 0
                    dialog_progress = SimpleDialogQt.ProgressDialog(
                        title="Synchronization, please wait",
                        content=("Synchronization of the file " + full_url + "\nThis operation may take a few minutes"))
                    dialog_progress.show()
                    if MetManagement.already_downloaded_compressed_server(
                            base_url, f"{target_subfolder}/{child_name}", local_filename_parsed
                    ):
                        print(f"{local_filename.resolve()} already downloaded -> skip")
                        return
                    local_filename.parent.mkdir(parents=True, exist_ok=True)
                    print(f"Downloading file: {child_url} to {local_filename.resolve()}")
                    process_file = subprocess.Popen(
                        ["curl", "-L","-#", "-C", "-o", str(local_filename.resolve()), child_url],
                        shell=True,text=True,stdout=subprocess.PIPE, stderr=subprocess.STDOUT,creationflags=creation_flags
                    )
                    if process_file.returncode != 0:
                        print(f"Direct download failed for {full_url}: {process.stderr}")
                    else:
                        print(f"File downloaded: {local_filename.resolve()}")
                        dialog_progress.stop()
        else:
            # No directory listing detected: treat it as a single file.
            local_filename = local_target_dir
            dialog_progress = SimpleDialogQt.ProgressDialog(
                title="Synchronization, please wait",
                content=("Synchronization of the file " + full_url + "\nThis operation may take a few minutes"))
            dialog_progress.show()
            if MetManagement.already_downloaded_compressed_server(
                    base_url, target_subfolder, str(local_filename.resolve())
            ):
                print(f"{local_filename.resolve()} already downloaded -> skip")
                return
            local_filename.parent.mkdir(parents=True, exist_ok=True)

            curl_command = f'curl -L -# -C - -o "{local_filename.resolve()}" "{full_url}"'
            process = subprocess.Popen(
                curl_command, shell=False, text=True,
                stdout=subprocess.PIPE, stderr=subprocess.STDOUT
            )
            if process.returncode != 0:
                print(f"Direct download failed for {full_url}: {process.stderr}")
            else:
                print(f"File downloaded: {local_filename.resolve()}")
                dialog_progress.stop()
            return


if __name__ == "__main__":
    # create the json needed to http / zipped stored
    in_repo_file="C:/modele_NLP/IFIA_models/repository.aait"
    create_index_file(in_repo_file)




import os
import sys

import Orange.data
from AnyQt.QtWidgets import QApplication, QPushButton
from Orange.data import StringVariable
from Orange.widgets.utils.signals import Input, Output
from AnyQt.QtWidgets import QCheckBox
from Orange.widgets.settings import Setting
if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils import base_widget
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
else:
    from orangecontrib.AAIT.utils import base_widget
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file


@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWFileWithPath(base_widget.BaseListWidget):
    name = "File with Path"
    category = "AAIT - TOOLBOX"
    description = "Load some tabular data specified with a filepath ('.../data/example.xlsx')."
    icon = "icons/owfilewithpath.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owfilewithpath.svg"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owfilewithpath.ui")
    want_control_area = False
    priority = 1060

    # Settings
    selected_column_name = Setting("path")

    class Inputs:
        filepath = Input("Path", str, auto_summary=False)
        path_table = Input("Path Table", Orange.data.Table)

    class Outputs:
        data = Output("Data", Orange.data.Table)

    strloadAsString: str = Setting('False')


    @Inputs.filepath
    def set_filepath(self, in_filepath):
        if in_filepath is not None:
            self.filepath = in_filepath
            self.run()

    @Inputs.path_table
    def set_path_table(self, in_path_table):
        self.in_path_table = in_path_table
        if self.in_path_table is None:
            self.Outputs.data.send(None)
            return
        if self.in_path_table is not None:
            self.var_selector.add_variables(self.in_path_table.domain)
            self.var_selector.select_variable_by_name(self.selected_column_name)

            self.filepath = in_path_table[0][self.selected_column_name].value
            self.run()

        if self.autorun:
            self.run()


    def __init__(self):
        super().__init__()
        # Qt Management
        #uic.loadUi(self.gui, self)
        self.setFixedWidth(500)
        self.setFixedHeight(450)
        self.checkbox_interface = self.findChild(QCheckBox, 'checkBox')
        if self.strloadAsString=="False":
            self.checkbox_interface.setChecked(False)
        else:
            self.checkbox_interface.setChecked(True)
        self.checkbox_interface.stateChanged.connect(self.on_checkbox_toggled)


        # Data Management
        self.filepath = None
        self.in_path_table = None
        self.data = None
        self.autorun = True
        self.post_initialized()

        self.pushButton_run = self.findChild(QPushButton, 'pushButton_send')
        self.pushButton_run.clicked.connect(self.run)
    def on_checkbox_toggled(self):
        if self.checkbox_interface.isChecked():
            self.strloadAsString="True"
        else:
            self.strloadAsString="False"

    def safe_repr_val(self,var, val, missing_as_empty=False):
        """
        Version robuste de var.repr_val(val) :
        - gère les valeurs NaN/Inf
        - gère les erreurs de datetime.fromtimestamp (OSError invalid argument)
        """
        try:
            if var.is_string:
                s = var.str_val(val)  # pas de guillemets
            else:
                s = var.repr_val(val)
            if missing_as_empty and s == "?":
                return ""
            return s
        except (ValueError, OverflowError, OSError):
            # cas des TimeVariable avec timestamp invalide
            return "" if missing_as_empty else "?"

    def load_table_as_strings(self,filepath: str) -> Orange.data.Table:
        """
        Charge un fichier Orange et renvoie un Table où toutes les colonnes
        (features, class, metas) sont converties en StringVariable et placées en metas.
        """
        missing_as_empty=False # si on passe a True on met "" à la place de ?
        t = Orange.data.Table.from_file(filepath)
        dom = t.domain

        # 2) Toutes les variables d'origine
        all_vars = list(dom.attributes) + list(dom.class_vars) + list(dom.metas)

        # 3) Créer des StringVariable correspondantes
        new_metas = tuple(Orange.data.StringVariable(v.name) for v in all_vars)

        # 4) Domaine final : que des metas
        new_domain = Orange.data.Domain([], None, new_metas)

        # 5) Créer table vide
        t_str = Orange.data.Table.from_domain(new_domain, len(t))

        # 6) Remplissage sécurisé
        for j, src_var in enumerate(all_vars):
            col_data = t.get_column(src_var)
            text_vals = [self.safe_repr_val(src_var, v, missing_as_empty) for v in col_data]
            t_str.metas[:, j] = text_vals

        # 7) Métadonnées
        t_str.name = t.name
        t_str.ids = t.ids
        t_str.attributes = dict(getattr(t, "attributes", {}))

        return t_str

    def run(self):
        self.error("")
        self.warning("")


        if self.filepath is None:
            self.Outputs.data.send(None)
            return

        self.filepath = self.filepath.strip('"')
        if not os.path.exists(self.filepath):
            self.error("error input file doesn t exist")
            self.Outputs.data.send(None)
            return
        if self.in_path_table is not None:
              # Verification of in_data
            if not self.selected_column_name in self.in_path_table.domain:
                self.warning(f'Previously selected column "{self.selected_column_name}" does not exist in your data.')
                self.Outputs.data.send(None)
                return

            if not isinstance(self.in_path_table.domain[self.selected_column_name], StringVariable):
                self.error('You must select a text variable.')
                self.Outputs.data.send(None)
                return

        if self.strloadAsString!="False":
            out_data = self.load_table_as_strings(self.filepath)
        else:
         out_data = Orange.data.Table.from_file(self.filepath)

        out_data.name = self.filepath
        self.Outputs.data.send(out_data)


    def post_initialized(self):
        pass

if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWFileWithPath()
    my_widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()



import sys

from AnyQt.QtWidgets import (QApplication, QDialog, QFileDialog, QLabel,QPushButton,
                             QMessageBox, QVBoxLayout,QSpinBox,QHBoxLayout)


from AnyQt import QtSvg,QtGui
def BoxInfo(text):
    """
    Open A simple info box
    """
    msg = QMessageBox()
    msg.setIcon(QMessageBox.Information)
    msg.setText(text)
    msg.setWindowTitle("Information")
    msg.setStandardButtons(QMessageBox.Ok)
    msg.exec()


def BoxWarning(text):
    """
    Open A simple warning box
    """
    msg = QMessageBox()
    msg.setIcon(QMessageBox.Warning)
    msg.setText(text)
    msg.setWindowTitle("Warning")
    msg.setStandardButtons(QMessageBox.Ok)
    msg.exec()


def BoxError(text):
    """
    Open A simple error box
    """
    msg = QMessageBox()
    msg.setIcon(QMessageBox.Critical)
    msg.setText(text)
    msg.setWindowTitle("Error")
    msg.setStandardButtons(QMessageBox.Ok)
    msg.exec()


def BoxSelectFolder(argself, default_path=None):
    """
    return "" if nothing was selected else the path
    """
    if default_path == None or default_path == "":
        folder = QFileDialog.getExistingDirectory()
    else:
        folder = QFileDialog.getExistingDirectory(argself, caption="Select a folder", directory=default_path)
    return folder.replace("\\", "/")

def BoxSelectExistingFile(argself,default_dir="",extention="All Files (*);;Text Files (*.txt)"):
    """
    return "" if nothing was selected else the path
    """
    try:  # Qt 6
        readonly = QFileDialog.Option.ReadOnly
    except AttributeError:  # Qt 5
        readonly = QFileDialog.ReadOnly

    fileName, _ = QFileDialog.getOpenFileName(
        argself, "Sélectionner un fichier", default_dir,
        options=readonly
    )

    if fileName:
        fileName=fileName.replace("\\", "/")
        return fileName
    else:
        return ""


def BoxYesNo(question):
    """
    return True if Yes is clicked, False in other cases
    """
    msg = QMessageBox()
    msg.setIcon(QMessageBox.Warning)
    msg.setText(question)
    msg.setStandardButtons(QMessageBox.No | QMessageBox.Yes)
    ret = msg.exec()
    if ret == msg.Yes:
        return True
    return False


class ProgressDialog(QDialog):
    """
    to use :
    dialog_progress = ProgressDialog(title="hello",content="blablalblagblal")
    dialog_progress.show()
    # do something
    dialog_progress.stop()
    """
    def __init__(self, title="Title",content="blablabla",parent=None):
        super().__init__(parent)
        self.setWindowTitle(title)
        self.setModal(True)

        layout = QVBoxLayout(self)
        self.label = QLabel(content, self)
        layout.addWidget(self.label)

    def show(self):
        super().show()
        QApplication.processEvents()

    def closeEvent(self, event):
        # ignore click on x
        event.ignore()

    def stop(self):
        self.accept()



def transformboutontools2wrench(bouttontool):
    """
        usage :
        self.bouttontool= self.findChild(QtWidgets.QToolButton, 'toolButton')
        SimpleDialogQt.transformboutontools2wrench(self.bouttontool)

    """
    wrench_svg = b"""<?xml version="1.0" encoding="iso-8859-1"?>
    <svg version="1.1" id="Capa_1" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" x="0px" y="0px"
    	 viewBox="0 0 303.848 303.848" style="enable-background:new 0 0 303.848 303.848;" xml:space="preserve">
    <g>
    	<path d="M297.023,52.375l-3.74-8.728l-50.361,50.348h-33.503l0.026-33.49l49.943-50.039l-9.023-3.618
    		c-10.129-4.055-20.739-6.105-31.543-6.105c-46.909,0-85.072,38.156-85.072,85.04c0,9.293,1.53,18.471,4.557,27.327L10.714,240.682
    		C3.798,247.584,0,256.775,0,266.543s3.805,18.953,10.714,25.842c6.896,6.915,16.08,10.72,25.849,10.72s18.946-3.805,25.849-10.72
    		L189.27,165.533c9.467,3.515,19.383,5.296,29.551,5.296c46.884,0,85.027-38.15,85.027-85.046
    		C303.848,74.214,301.553,62.98,297.023,52.375z M218.821,157.975c-9.891,0-19.493-1.986-28.542-5.887l-4.004-1.735L53.323,283.298
    		c-8.965,8.965-24.576,8.965-33.529,0c-4.48-4.467-6.947-10.431-6.947-16.755c0-6.337,2.468-12.288,6.947-16.768l133.492-133.472
    		l-1.555-3.927c-3.406-8.573-5.135-17.526-5.135-26.594c0-39.801,32.404-72.186,72.218-72.186c5.81,0,11.575,0.701,17.198,2.095
    		L196.584,55.19l-0.039,51.665h51.691l40.193-40.181c1.71,6.195,2.558,12.59,2.558,19.113
    		C290.994,125.59,258.622,157.975,218.821,157.975z M46.061,265.689c0,5.135-4.165,9.3-9.3,9.3c-5.141,0-9.306-4.165-9.306-9.3
    		c0-5.148,4.165-9.312,9.306-9.312C41.89,256.376,46.061,260.541,46.061,265.689z"/>
    </g>
    <g>
    </g>
    <g>
    </g>
    <g>
    </g>
    <g>
    </g>
    <g>
    </g>
    <g>
    </g>
    <g>
    </g>
    <g>
    </g>
    <g>
    </g>
    <g>
    </g>
    <g>
    </g>
    <g>
    </g>
    <g>
    </g>
    <g>
    </g>
    <g>
    </g>
    </svg>
    """
    renderer = QtSvg.QSvgRenderer(wrench_svg)
    pixmap = QtGui.QPixmap(24, 24)  # pixmal size
    pixmap.fill(QtGui.QColor("transparent"))  # Fond transparent
    painter = QtGui.QPainter(pixmap)
    renderer.render(painter)
    painter.end()
    bouttontool.setIconSize(pixmap.size())  # Ajuste la taille de l'icône
    bouttontool.setFixedSize(pixmap.size())  # Ajuste la taille du bouton
    # Appliquer l'icône au bouton
    bouttontool.setIcon(QtGui.QIcon(pixmap))




class NumberInputDialog(QDialog):
    """
    usage
    selected_value = get_number_from_dialog("select a number between 1 et 100 :", 1, 100)
    print("result :", selected_value)
    """
    def __init__(self, text, value_min, value_max, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Select a Number")

        layout = QVBoxLayout(self)

        # Ajout du texte descriptif
        self.label = QLabel(text, self)
        layout.addWidget(self.label)

        # Ajout du spinbox
        self.spinbox = QSpinBox(self)
        self.spinbox.setMinimum(value_min)
        self.spinbox.setMaximum(value_max)
        layout.addWidget(self.spinbox)

        # Boutons OK et Annuler
        button_layout = QHBoxLayout()
        self.ok_button = QPushButton("OK", self)
        self.cancel_button = QPushButton("Cancel", self)

        button_layout.addWidget(self.ok_button)
        button_layout.addWidget(self.cancel_button)
        layout.addLayout(button_layout)

        # Connexions des boutons
        self.ok_button.clicked.connect(self.accept)
        self.cancel_button.clicked.connect(self.reject)

    def get_value(self):
        return self.spinbox.value()

def get_number_from_dialog(text, value_min, value_max):
    """
    usage
    selected_value = get_number_from_dialog("select a number between 1 et 100 :", 1, 100)
    print("result :", selected_value)
    """
    ## surement fausse alarme de pyflakes
    _ = QApplication(sys.argv) if not QApplication.instance() else QApplication.instance()
    dialog = NumberInputDialog(text, value_min, value_max)
    if dialog.exec():
        return dialog.get_value()
    return None



import os
import sys
import re
from typing import Optional,Dict,List,Tuple
from AnyQt.QtWidgets import QTableWidgetItem

from Orange.widgets.settings import Setting
import Orange.data
from Orange.data import Table
from AnyQt.QtWidgets import QApplication
from Orange.widgets import widget

# Import des bibliothèques Word
from docx import Document
from docx.oxml.ns import qn
from docx.oxml import OxmlElement
if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic

else:
    from orangecontrib.AAIT.utils.import_uic import uic

class OWDocumentGenerator(widget.OWWidget):
    """
    Widget simplifié pour générer des documents Word
    - Input 1 : Chemins (template in/out)
    - Input 2 : Remplacements texte (balise -> valeur)
    - Input 3 : Tables à insérer (nom = balise cible)
    """
    name = "Document Generator"
    description = "Modify Word documents with replacements and tables."
    category = "AAIT - TOOLBOX"
    icon = "icons/document_generator.png"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/document_generator.png"

    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owgenerate_word.ui")
    want_control_area = False
    priority = 1070

    key_remplace: str = Setting("True")# "True" => replace [col] with row0 value else   => insert full table at [TableKey]


    class Inputs:
        data = Orange.widgets.widget.Input("Data", Orange.data.Table)

    class Outputs:
        data = Orange.widgets.widget.Output("Data", Orange.data.Table)

    # -----------------------------
    # Colors: %!color!% prefix (headers + cells)
    # -----------------------------
    COLOR_MAP = {
        # neutres
        "blanc": "FFFFFF", "white": "FFFFFF",
        "gris": "E0E0E0", "gray": "E0E0E0", "grey": "E0E0E0",
        "gris_clair": "F2F2F2",

        # pastel lisibles
        "jaune": "FFF2CC", "yellow": "FFF2CC",
        "vert": "E2F0D9", "green": "E2F0D9",
        "rouge": "FCE4D6", "red": "FCE4D6",
        "bleu": "DDEBF7", "blue": "DDEBF7",
        "violet": "EDE7F6", "purple": "EDE7F6",
        "orange": "FBE5D6",
        "cyan": "E7F3F8", "turquoise": "E7F3F8",
        "rose": "FCE4EC", "pink": "FCE4EC",
    }

    def __init__(self):
        super().__init__()
        self.setFixedWidth(470)
        self.setFixedHeight(300)
        uic.loadUi(self.gui, self)

        self.tableWidget.insertRow(self.tableWidget.rowCount())


        self.tableWidget.setItem(0, 0, QTableWidgetItem("C:/path.docx"))
        self.tableWidget.setItem(0, 1, QTableWidgetItem("Value1"))
        self.tableWidget.setItem(0, 2, QTableWidgetItem("Value2"))
        self.tableWidget.setItem(0, 3, QTableWidgetItem("Value3"))

        self.tableWidget_2.insertRow(self.tableWidget_2.rowCount())

        self.tableWidget_2.setItem(0, 0, QTableWidgetItem("C:/path.docx"))
        self.tableWidget_2.setItem(0, 1, QTableWidgetItem("Value1"))
        self.tableWidget_2.setItem(0, 2, QTableWidgetItem("%!red!%..."))
        self.tableWidget_2.setItem(0, 3, QTableWidgetItem("..."))




        self.data: Optional[Table] = None

        # Word table style (optional)
        self.table_style_name = "Table Grid"
        if self.key_remplace=="True":
            self.radioButton.setChecked(True)
        else:
            self.radioButton_2.setChecked(True)


        self.radioButton.toggled.connect(self.on_radio_changed)
        self.radioButton_2.toggled.connect(self.on_radio_changed)




    def on_radio_changed(self, checked):
        if not checked:
            return  # on ignore le "décoché"

        sender = self.sender()

        if sender == self.radioButton:
            self.key_remplace="True"
        elif sender == self.radioButton_2:
            self.key_remplace="False"

    # -----------------------------
    # Input
    # -----------------------------
    @Inputs.data
    def set_data(self, data: Optional[Table]):
        self.data = data
        self.process()

    # -----------------------------
    # Helpers
    # -----------------------------
    def _clean_orange_value(self, s) -> str:
        if s is None:
            return ""
        v = str(s).strip()
        if v in ("?", "nan", "NaN", "None"):
            return ""
        return v

    def _normalize_hex(self, color: str) -> str:
        if not color:
            return "FFFFFF"
        c = str(color).strip().lower()
        if c.startswith("#"):
            c = c[1:]
        if re.fullmatch(r"[0-9a-f]{6}", c):
            return c.upper()
        return self.COLOR_MAP.get(c, "FFFFFF")

    def parse_color_prefix(self, value: str) -> Tuple[str, Optional[str]]:
        """
        Detect prefix %!color!% or %!#RRGGBB!%
        Returns (text_without_prefix, hex_color_or_None)
        """
        if value is None:
            return "", None

        s = str(value)
        m = re.match(r"^%!(.*?)!%", s)
        if not m:
            return s, None

        token = m.group(1).strip()
        text = s[m.end():]
        return text, self._normalize_hex(token)

    def set_cell_bg(self, cell, hex_rgb: str):
        hex_rgb = self._normalize_hex(hex_rgb)

        tcPr = cell._tc.get_or_add_tcPr()
        shd = tcPr.find(qn("w:shd"))
        if shd is None:
            shd = OxmlElement("w:shd")
            tcPr.append(shd)

        shd.set(qn("w:val"), "clear")
        shd.set(qn("w:color"), "auto")
        shd.set(qn("w:fill"), hex_rgb)

    def _get_first_row_mapping(self, table: Table) -> Dict[str, str]:
        if table is None or len(table) == 0:
            return {}

        row0 = table[0]
        mapping: Dict[str, str] = {}

        for var in table.domain.attributes:
            try:
                mapping[var.name] = self._clean_orange_value(row0[var])
            except Exception:
                mapping[var.name] = ""

        for var in table.domain.class_vars:
            try:
                mapping[var.name] = self._clean_orange_value(row0[var])
            except Exception:
                mapping[var.name] = ""

        for j, var in enumerate(table.domain.metas):
            try:
                mapping[var.name] = self._clean_orange_value(table.metas[0, j])
            except Exception:
                mapping[var.name] = ""

        return mapping

    def _bracketed_only(self, key: str) -> str:
        k = (key or "").strip()
        if not k:
            return ""
        if k.startswith("[") and k.endswith("]") and len(k) >= 2:
            inner = k[1:-1].strip()
            return f"[{inner}]" if inner else ""
        return f"[{k}]"

    def _extract_full_table_matrix(self, table: Table, exclude_names: set) -> Tuple[List[str], List[List[str]]]:
        """
        Extract headers + rows from Orange Table (all rows),
        using attrs + targets + metas, excluding certain names.
        NOTE: headers keep raw names (can include %!color!% prefix).
        """
        headers: List[str] = []
        specs = []  # list of tuples describing how to fetch values (kind, var, meta_index)

        for var in table.domain.attributes:
            if var.name not in exclude_names:
                headers.append(var.name)
                specs.append(("attr", var, None))

        for var in table.domain.class_vars:
            if var.name not in exclude_names:
                headers.append(var.name)
                specs.append(("target", var, None))

        for j, var in enumerate(table.domain.metas):
            if var.name not in exclude_names:
                headers.append(var.name)
                specs.append(("meta", None, j))

        rows: List[List[str]] = []
        for i in range(len(table)):
            row_vals: List[str] = []
            for kind, var, j in specs:
                try:
                    if kind in ("attr", "target"):
                        row_vals.append(self._clean_orange_value(table[i][var]))
                    else:
                        row_vals.append(self._clean_orange_value(table.metas[i, j]))
                except Exception:
                    row_vals.append("")
            rows.append(row_vals)

        return headers, rows

    # -----------------------------
    # Format-preserving replacement in paragraph runs
    # -----------------------------
    def _replace_in_paragraph_preserve_format(self, paragraph, placeholder: str, replacement: str) -> int:
        if not placeholder:
            return 0

        runs = paragraph.runs
        if not runs:
            return 0

        full = "".join(r.text for r in runs)
        if placeholder not in full:
            return 0

        repl_count = 0

        while True:
            runs = paragraph.runs
            full = "".join(r.text for r in runs)

            start = full.find(placeholder)
            if start < 0:
                break
            end = start + len(placeholder)

            spans = []
            pos = 0
            for idx, r in enumerate(runs):
                t = r.text or ""
                spans.append((idx, pos, pos + len(t)))
                pos += len(t)

            start_run = None
            for (i, a, b) in spans:
                if a <= start < b:
                    start_run = (i, a, b)
                    break
            if start_run is None:
                break

            end_run = None
            for (i, a, b) in spans:
                if a < end <= b:
                    end_run = (i, a, b)
                    break
            if end_run is None:
                for (i, a, b) in spans:
                    if end == a:
                        end_run = (i, a, b)
                        break

            si, sa, sb = start_run
            if end_run is None:
                ei, ea, eb = si, sa, sb
            else:
                ei, ea, eb = end_run

            start_off = start - sa

            if si == ei:
                r = runs[si]
                text = r.text or ""
                before = text[:start_off]
                after = text[start_off + len(placeholder):]
                r.text = before + replacement + after
                repl_count += 1
                continue

            first = runs[si]
            last = runs[ei]

            first_text = first.text or ""
            last_text = last.text or ""

            before = first_text[:start_off]
            after = last_text[end - ea:]

            first.text = before

            insert_run = paragraph.add_run(replacement)
            insert_run.bold = first.bold
            insert_run.italic = first.italic
            insert_run.underline = first.underline
            insert_run.style = first.style

            try:
                if first.font is not None:
                    insert_run.font.name = first.font.name
                    insert_run.font.size = first.font.size
                    if getattr(first.font.color, "rgb", None) is not None:
                        insert_run.font.color.rgb = first.font.color.rgb
                    insert_run.font.highlight_color = first.font.highlight_color
                    insert_run.font.all_caps = first.font.all_caps
                    insert_run.font.small_caps = first.font.small_caps
                    insert_run.font.strike = first.font.strike
                    insert_run.font.double_strike = first.font.double_strike
                    insert_run.font.subscript = first.font.subscript
                    insert_run.font.superscript = first.font.superscript
            except Exception:
                pass

            first._element.addnext(insert_run._element)

            for idx in range(si + 1, ei + 1):
                runs[idx].text = ""

            runs[ei].text = after

            repl_count += 1

        return repl_count

    # -----------------------------
    # Replace everywhere (body + tables + headers/footers)
    # -----------------------------
    def _replace_everywhere_in_container(self, container, placeholder: str, replacement: str) -> int:
        count = 0

        for p in getattr(container, "paragraphs", []):
            count += self._replace_in_paragraph_preserve_format(p, placeholder, replacement)

        for t in getattr(container, "tables", []):
            for row in t.rows:
                for cell in row.cells:
                    count += self._replace_everywhere_in_container(cell, placeholder, replacement)

        return count

    def _replace_everywhere(self, doc: Document, placeholder: str, replacement: str) -> int:
        total = 0
        total += self._replace_everywhere_in_container(doc, placeholder, replacement)

        for section in doc.sections:
            if section.header is not None:
                total += self._replace_everywhere_in_container(section.header, placeholder, replacement)
            if section.footer is not None:
                total += self._replace_everywhere_in_container(section.footer, placeholder, replacement)

        return total

    # -----------------------------
    # Insert Word table at a [Tag] (body + headers/footers)
    # -----------------------------
    def _apply_table_style_if_possible(self, table_obj) -> None:
        try:
            if self.table_style_name:
                table_obj.style = self.table_style_name
        except Exception:
            pass

    def _insert_table_after_paragraph(self, doc: Document, paragraph, headers: List[str], rows: List[List[str]]) -> None:
        n_cols = max(1, len(headers))
        n_rows = 1 + len(rows)

        tbl = doc.add_table(rows=n_rows, cols=n_cols)
        self._apply_table_style_if_possible(tbl)

        # Header row (support %!color!% prefix)
        for j, raw_h in enumerate(headers):
            cell_h = tbl.cell(0, j)
            clean_h, hex_bg = self.parse_color_prefix(raw_h)
            cell_h.text = clean_h
            if hex_bg:
                self.set_cell_bg(cell_h, hex_bg)

        # Data rows (support %!color!% prefix per cell)
        for i, row_vals in enumerate(rows, start=1):
            for j in range(n_cols):
                v = row_vals[j] if j < len(row_vals) else ""
                cell_d = tbl.cell(i, j)

                clean_v, hex_bg = self.parse_color_prefix(v)
                cell_d.text = "" if clean_v is None else str(clean_v)

                if hex_bg:
                    self.set_cell_bg(cell_d, hex_bg)

        paragraph._element.addnext(tbl._element)

    def _insert_table_at_tag_in_container(self, doc: Document, container, tag: str,
                                         headers: List[str], rows: List[List[str]]) -> bool:
        for p in getattr(container, "paragraphs", []):
            try:
                if tag in (p.text or ""):
                    self._replace_in_paragraph_preserve_format(p, tag, "")
                    self._insert_table_after_paragraph(doc, p, headers, rows)
                    return True
            except Exception as e:
                import traceback
                print("Insert table error:", e)
                print(traceback.format_exc())
                continue

        for t in getattr(container, "tables", []):
            for row in t.rows:
                for cell in row.cells:
                    if self._insert_table_at_tag_in_container(doc, cell, tag, headers, rows):
                        return True

        return False

    def _insert_table_at_tag(self, doc: Document, tag: str,
                             headers: List[str], rows: List[List[str]]) -> bool:
        if self._insert_table_at_tag_in_container(doc, doc, tag, headers, rows):
            return True

        for section in doc.sections:
            if section.header is not None:
                if self._insert_table_at_tag_in_container(doc, section.header, tag, headers, rows):
                    return True
            if section.footer is not None:
                if self._insert_table_at_tag_in_container(doc, section.footer, tag, headers, rows):
                    return True

        return False

    # -----------------------------
    # Main
    # -----------------------------
    def process(self):
        self.error("")
        self.warning("")

        if self.data is None:
            self.Outputs.data.send(None)
            return

        if len(self.data) == 0:
            self.error("La table d'entrée est vide (aucune ligne).")
            self.Outputs.data.send(None)
            return

        try:
            mapping = self._get_first_row_mapping(self.data)

            if "DocxPath" not in mapping:
                self.error("Colonne 'DocxPath' introuvable (attributes/targets/metas).")
                self.Outputs.data.send(None)
                return

            docx_path = self._clean_orange_value(mapping.get("DocxPath", "")).strip()
            if not docx_path:
                self.error("La valeur de 'DocxPath' est vide sur la première ligne.")
                self.Outputs.data.send(None)
                return

            docx_path = docx_path.replace("\\", os.sep).replace("/", os.sep)
            if not os.path.exists(docx_path):
                self.error(f"Fichier DOCX introuvable : {docx_path}")
                self.Outputs.data.send(None)
                return

            doc = Document(docx_path)

            # MODE 1: key/value replacements in [tags] from row0
            if str(self.key_remplace) == "True":
                replacements = {k: v for k, v in mapping.items() if k != "DocxPath"}
                for key, val in replacements.items():
                    tag = self._bracketed_only(key)  # only [key]
                    if not tag:
                        continue
                    rep = self._clean_orange_value(val)
                    self._replace_everywhere(doc, tag, rep)

                doc.save(docx_path)
                self.Outputs.data.send(self.data)
                return

            # MODE 2: insert full table at [TableKey]
            if "TableKey" not in mapping:
                self.error("Mode table: colonne 'TableKey' introuvable sur la première ligne.")
                self.Outputs.data.send(None)
                return

            table_key = self._clean_orange_value(mapping.get("TableKey", "")).strip()
            table_key, _ = self.parse_color_prefix(table_key)  # enlève %!couleur!%
            table_key = (table_key or "").strip()
            if not table_key:
                self.error("Mode table: la valeur de 'TableKey' est vide sur la première ligne.")
                self.Outputs.data.send(None)
                return

            insert_tag = self._bracketed_only(table_key)  # => [Tableau1]
            if not insert_tag:
                self.error("Mode table: TableKey invalide.")
                self.Outputs.data.send(None)
                return
            exclude = {"DocxPath", "TableKey"}
            headers, rows = self._extract_full_table_matrix(self.data, exclude_names=exclude)

            if not headers:
                self.error("Mode table: aucune colonne à insérer (tout est exclu).")
                self.Outputs.data.send(None)
                return

            inserted = self._insert_table_at_tag(doc, insert_tag, headers, rows)
            if not inserted:
                self.error(f"Balise d'insertion non trouvée dans le DOCX : {insert_tag}")
                self.Outputs.data.send(None)
                return

            doc.save(docx_path)
            self.Outputs.data.send(self.data)

        except Exception as e:
            self.error(f"Erreur : {e}")
            self.Outputs.data.send(None)

if __name__ == "__main__":
    app = QApplication(sys.argv)
    widget = OWDocumentGenerator()
    widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()



# wordchunker_deprecated.py
# -----------------------------------------------------------------------------
# Compatibility shim reproducing chonkie==0.4.1 WordChunker behavior (including
# historical quirks/bugs), while being usable inside chonkie>=1.x pipelines.
#
# What you get:
# - A local WordChunker class with the same logic as chonkie 0.4.1 word chunker.
# - Preserves newlines: words are spans like r"(\s*\S+)" joined with "".
# - chunk_size / chunk_overlap are TOKEN budgets (not "word" counts), measured
#   via the provided tokenizer backend.
# - Reproduces 0.4.1 overlap-loop quirk: iterates range(previous_chunk_length)
#   where previous_chunk_length is a TOKEN count (not word count).
# - Reproduces 0.4.1 final-chunk quirk: _create_chunk called without current_index
#   (defaults to 0), which can yield start_index via .find from the start.
#
# Added for practicality in your AAIT / chonkie>=1.5.2 integration:
# - WordChunker is callable: chunker(text) == chunker.chunk(text)
# - Accepts tokenizer="character" (character-count pseudo tokenizer)
# - Accepts a SentenceTransformer instance as tokenizer (uses .tokenizer underneath)
#
# NOTE (important bugfix vs earlier draft):
# Many HuggingFace tokenizers are *callable* (tokenizer(text) -> BatchEncoding).
# We must NOT mis-detect them as a generic callable token counter. We detect
# transformers/tokenizers backends *first* by attributes, then fall back to callable.
# -----------------------------------------------------------------------------

from __future__ import annotations

import importlib
import inspect
import re
from dataclasses import dataclass
from typing import Any, Callable, List, Union


@dataclass(frozen=True)
class Chunk:
    text: str
    start_index: int
    end_index: int
    token_count: int


# -----------------------------
# Helpers for robust tokenizers
# -----------------------------

def _unwrap_possible_sentence_transformer(obj: Any) -> Any:
    """
    If `obj` looks like a SentenceTransformer, return its underlying HF tokenizer
    when available. SentenceTransformer usually has `.tokenizer`.
    """
    if obj is None:
        return obj
    try:
        if hasattr(obj, "tokenizer") and "SentenceTransformer" in type(obj).__name__:
            tok = getattr(obj, "tokenizer", None)
            if tok is not None:
                return tok
    except Exception:
        pass
    return obj


def _make_character_tokenizer() -> Any:
    """
    Special compatibility: when user passes tokenizer="character", count tokens
    as characters, and support batch encoding.
    """
    class _CharTokenizer:
        def encode(self, text: str):
            return list(range(len(text)))

        def encode_batch(self, texts: List[str]):
            return [self.encode(t) for t in texts]

    return _CharTokenizer()


# -----------------------------
# Minimal BaseChunker (0.4.1-ish)
# -----------------------------

class BaseChunker:
    """
    Minimal subset of chonkie.chunker.base.BaseChunker needed by WordChunker 0.4.1
    """

    def __init__(self, tokenizer_or_token_counter: Union[str, Any, Callable[[str], int]]):
        tokenizer_or_token_counter = _unwrap_possible_sentence_transformer(tokenizer_or_token_counter)

        if tokenizer_or_token_counter == "character":
            tokenizer_or_token_counter = _make_character_tokenizer()

        if isinstance(tokenizer_or_token_counter, str):
            self.tokenizer = self._load_tokenizer(tokenizer_or_token_counter)
        else:
            self.tokenizer = tokenizer_or_token_counter

        self._tokenizer_backend = self._get_tokenizer_backend()
        self.token_counter = self._get_tokenizer_counter()

    def _get_tokenizer_backend(self) -> str:
        t = self.tokenizer
        tname = type(t).__name__
        ttype = str(type(t))

        # 1) transformers-style tokenizer (callable, returns BatchEncoding, has encode, often batch_encode_plus)
        if hasattr(t, "batch_encode_plus") or "transformers" in ttype or "PreTrainedTokenizer" in tname:
            return "transformers"

        # 2) tokenizers rust-style (has encode_batch with add_special_tokens)
        if "tokenizers" in ttype:
            return "tokenizers"

        # 3) tiktoken encodings
        if "tiktoken" in ttype:
            return "tiktoken"

        # 4) our custom / other encoders that implement encode_batch
        if hasattr(t, "encode_batch"):
            return "encode_batch"

        # 5) basic encode-only objects
        if hasattr(t, "encode"):
            return "encode_only"

        # 6) finally: a callable token *counter* function: f(text)->int
        # (must be LAST so we don't mis-detect HF tokenizers as callable counters)
        if callable(t) or inspect.isfunction(t) or inspect.ismethod(t):
            return "callable_counter"

        raise ValueError(f"Tokenizer backend {ttype} not supported")

    def _load_tokenizer(self, tokenizer_name: str):
        # Same overall strategy as 0.4.1: try tiktoken -> autotiktokenizer -> tokenizers -> transformers
        try:
            if importlib.util.find_spec("tiktoken") is not None:
                from tiktoken import get_encoding
                return get_encoding(tokenizer_name)
            raise RuntimeError("tiktoken not available")
        except Exception:
            try:
                if importlib.util.find_spec("autotiktokenizer") is not None:
                    from autotiktokenizer import AutoTikTokenizer
                    return AutoTikTokenizer.from_pretrained(tokenizer_name)
                raise RuntimeError("autotiktokenizer not available")
            except Exception:
                try:
                    if importlib.util.find_spec("tokenizers") is not None:
                        from tokenizers import Tokenizer
                        return Tokenizer.from_pretrained(tokenizer_name)
                    raise RuntimeError("tokenizers not available")
                except Exception:
                    if importlib.util.find_spec("transformers") is not None:
                        from transformers import AutoTokenizer
                        return AutoTokenizer.from_pretrained(tokenizer_name)
                    raise ValueError(
                        "Tokenizer not found in: transformers, tokenizers, autotiktokenizer, tiktoken. "
                        "Install one of these."
                    )

    def _get_tokenizer_counter(self) -> Callable[[str], int]:
        t = self.tokenizer
        if self._tokenizer_backend == "transformers":
            return lambda text: len(t.encode(text, add_special_tokens=False))
        if self._tokenizer_backend == "tokenizers":
            return lambda text: len(t.encode(text, add_special_tokens=False).ids)
        if self._tokenizer_backend == "tiktoken":
            return lambda text: len(t.encode(text))
        if self._tokenizer_backend == "encode_batch":
            if hasattr(t, "encode"):
                return lambda text: len(t.encode(text))
            return lambda text: len(t.encode_batch([text])[0])
        if self._tokenizer_backend == "encode_only":
            return lambda text: len(t.encode(text))
        if self._tokenizer_backend == "callable_counter":
            return t  # type: ignore[return-value]
        raise ValueError("Tokenizer backend not supported for token counting")

    def _encode_batch(self, texts: List[str]) -> List[List[int]]:
        """
        Return list of token-id lists. Only lengths are used by WordChunker.
        """
        t = self.tokenizer
        if self._tokenizer_backend == "transformers":
            # batch_encode_plus exists on most HF tokenizers; if not, fall back to __call__
            if hasattr(t, "batch_encode_plus"):
                return t.batch_encode_plus(texts, add_special_tokens=False)["input_ids"]
            # Fallback: tokenizer(texts, add_special_tokens=False) -> BatchEncoding with input_ids
            enc = t(texts, add_special_tokens=False)
            return enc["input_ids"]
        if self._tokenizer_backend == "tokenizers":
            return [e.ids for e in t.encode_batch(texts, add_special_tokens=False)]
        if self._tokenizer_backend == "tiktoken":
            return t.encode_batch(texts)
        if self._tokenizer_backend == "encode_batch":
            return t.encode_batch(texts)
        if self._tokenizer_backend == "encode_only":
            return [t.encode(x) for x in texts]
        if self._tokenizer_backend == "callable_counter":
            # emulate "ids" with dummy list of length == token_count
            out: List[List[int]] = []
            for x in texts:
                n = int(t(x))
                out.append(list(range(n)))
            return out
        raise ValueError(f"Tokenizer backend {self._tokenizer_backend} not supported.")


# -----------------------------
# WordChunker (exact 0.4.1 logic)
# -----------------------------

class WordChunker(BaseChunker):
    """
    Exact port of chonkie==0.4.1 WordChunker (chunker/word.py), with identical behavior/quirks.
    """

    def __init__(
        self,
        tokenizer: Union[str, Any] = "gpt2",
        chunk_size: int = 512,
        chunk_overlap: int = 128,
    ):
        super().__init__(tokenizer)

        if chunk_size <= 0:
            raise ValueError("chunk_size must be positive")
        if chunk_overlap >= chunk_size:
            raise ValueError("chunk_overlap must be less than chunk_size")

        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap

    def _split_into_words(self, text: str) -> List[str]:
        split_points = [match.end() for match in re.finditer(r"(\s*\S+)", text)]
        words: List[str] = []
        prev = 0

        for point in split_points:
            words.append(text[prev:point])
            prev = point

        if prev < len(text):
            words.append(text[prev:])

        return words

    def _create_chunk(
        self,
        words: List[str],
        text: str,
        token_count: int,
        current_index: int = 0,
    ) -> Chunk:
        chunk_text = "".join(words)
        start_index = text.find(chunk_text, current_index)
        return Chunk(
            text=chunk_text,
            start_index=start_index,
            end_index=start_index + len(chunk_text),
            token_count=token_count,
        )

    def _get_word_list_token_counts(self, words: List[str]) -> List[int]:
        words = [word for word in words if word != ""]
        encodings = self._encode_batch(words)
        return [len(encoding) for encoding in encodings]

    def chunk(self, text: str) -> List[Chunk]:
        if not text or not text.strip():
            return []

        words = self._split_into_words(text)
        lengths = self._get_word_list_token_counts(words)
        chunks: List[Chunk] = []

        current_chunk: List[str] = []
        current_chunk_length = 0
        current_index = 0

        for i, (word, length) in enumerate(zip(words, lengths)):
            if current_chunk_length + length <= self.chunk_size:
                current_chunk.append(word)
                current_chunk_length += length
            else:
                chunk = self._create_chunk(current_chunk, text, current_chunk_length, current_index)
                chunks.append(chunk)

                previous_chunk_length = current_chunk_length
                current_index = chunk.end_index

                overlap: List[str] = []
                overlap_length = 0

                # Quirk/bug-compatible loop: previous_chunk_length is token count
                for j in range(0, previous_chunk_length):
                    cwi = i - 1 - j
                    if cwi < 0:
                        break
                    oword = words[cwi]
                    olength = lengths[cwi]
                    if overlap_length + olength <= self.chunk_overlap:
                        overlap.append(oword)
                        overlap_length += olength
                    else:
                        break

                current_chunk = [w for w in reversed(overlap)]
                current_chunk_length = overlap_length

                current_chunk.append(word)
                current_chunk_length += length

        if current_chunk:
            # Quirk/bug-compatible: current_index not passed (defaults to 0)
            chunk = self._create_chunk(current_chunk, text, current_chunk_length)
            chunks.append(chunk)

        return chunks

    def __call__(self, text: str) -> List[Chunk]:
        return self.chunk(text)

    def __repr__(self) -> str:
        return f"WordChunker(chunk_size={self.chunk_size}, chunk_overlap={self.chunk_overlap})"


def chunk_words(content: str, tokenizer: Any, chunk_size: int = 300, chunk_overlap: int = 100):
    chunker = WordChunker(tokenizer=tokenizer, chunk_size=chunk_size, chunk_overlap=chunk_overlap)
    chunks = chunker.chunk(content)
    return [c.text for c in chunks], []



import sys
from Orange.widgets.settings import Setting
from Orange.widgets import widget
from Orange.widgets.utils.signals import  Output
from AnyQt.QtWidgets import QApplication,QMainWindow
from AnyQt.QtCore import QTimer
import os
if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
else:
    from orangecontrib.AAIT.utils.import_uic import uic

class OWQuadrantclicker(widget.OWWidget):
    name = "Quadrant Clicker"
    description = "Choose which quadrant widgets are displayed in at startup"
    category = "AAIT - TOOLBOX"
    icon = "icons/quadrantclicker.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/quadrantclicker.svg"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owquadrant_clicker.ui")
    want_control_area = False
    priority = 1089

    str_checkbox_statut: str = Setting("False")

    class Outputs:
        output_autoshow  = Output("AutoShowConfiguration", str, auto_summary=False)

    def minimize_all_qmainwindows(self):
        for w in QApplication.topLevelWidgets():
            try:
                if isinstance(w, QMainWindow):
                    w.showMinimized()
            except Exception as e:
                print(e)

    def __init__(self):
        super().__init__()
        uic.loadUi(self.gui, self)
        if self.str_checkbox_statut=="False":
            self.checkBox.setChecked(False)
        else:
            self.checkBox.setChecked(True)

        self.btn_tl.clicked.connect(lambda: self._p("top-left"))
        self.btn_tr.clicked.connect(lambda: self._p("top-right"))
        self.btn_bl.clicked.connect(lambda: self._p("bottom-left"))
        self.btn_br.clicked.connect(lambda: self._p("bottom-right"))

        # # Edges
        self.btn_top.clicked.connect(lambda: self._p("top"))
        self.btn_bottom.clicked.connect(lambda: self._p("bottom"))
        self.btn_left.clicked.connect(lambda: self._p("left"))
        self.btn_right.clicked.connect(lambda: self._p("right"))
        #
        # # Bottom action buttons
        self.btn_fs.clicked.connect(lambda: self._p("fullscreen"))
        self.btn_disabled.clicked.connect(lambda: self._p("none"))

        self.pushButton_free.clicked.connect(self.send_free)


        self.checkBox.toggled.connect(self.on_checkbox_toggled)

        if self.str_checkbox_statut != "False":
            QTimer.singleShot(50, self.minimize_all_qmainwindows)

    def send_free(self):
        delta_h=int(self.spinBox_dh.value())
        if delta_h+int(self.spinBox_h.value())>100:
            delta_h=100-int(self.spinBox_h.value())

        delta_l=int(self.spinBox_dl.value())
        if delta_l+int(self.spinBox_l.value())>100:
            delta_l=100-int(self.spinBox_l.value())


        str_value=f"v{str(self.spinBox_h.value())}_dv{str(delta_h)}_h{str(self.spinBox_l.value())}_dh{str(delta_l)}"
        self.label.setText("Selected position : "+str(str_value))
        self.Outputs.output_autoshow.send(str(str_value))



    def _p(self, s: str):
        self.label.setText("Selected position : "+str(s))
        self.Outputs.output_autoshow.send(str(s))

    def on_checkbox_toggled(self, checked: bool):
        if checked:
            self.str_checkbox_statut = "True"
        else:
            self.str_checkbox_statut = "False"





if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWQuadrantclicker()
    my_widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()

